from artemis.fileman.local_dir import get_artemis_data_path
import os
import pickle
import numpy as np
from six.moves import input
from artemis.general.hashing import fixed_hash_eq


def assert_variable_matches_between_runs(var, var_name, show='max_diff'):
    """
    This function allows you to check that a variable is equal between different runs of your program.  It is useful for
    rooting out unexpected randomness in your code.

    For example, suppose you have a variable my_var that for some reason is not identical between runs - but its value
    diverges only after some number of iterations.

    In your loop

        while True:

            ....
            assert_variable_matches_between_runs(my_var, 'my_var')

    :param var: A variable value (preferably a python primative or numpy array)
    :param var_name: A name to identify your variable
    """

    variable_matches_between_runs(var=var, var_name=var_name, show=show, raise_error=True)


def variable_matches_between_runs(var, var_name, raise_error = False, show='max_diff'):
    """
    This function allows you to check that a variable is equal between different runs of your program.  It is useful for
    rooting out unexpected randomness in your code.

    For example, suppose you have a variable my_var that for some reason is not identical between runs - but its value
    diverges only after some number of iterations.

    In your loop

        while True:

            ....
            assert_variable_matches_between_runs(my_var, 'my_var')

    :param var: A variable value (preferably a python primative or numpy array)
    :param var_name: A name to identify your variable
    """

    if var_name not in _var_write_status:
        _var_write_status[var_name] = not os.path.exists(_get_tracker_file(var_name))

    if _var_write_status[var_name]:
        _write_next_var_value(var, var_name)
        return None
    else:
        if var_name not in _var_calls:
            _var_calls[var_name] = 0
        _var_calls[var_name] += 1
        loaded_var = _get_next_saved_var_value(var_name)
        if not fixed_hash_eq(var, loaded_var):
            if show=='full':
                message = '{} @ {} does not match saved value: \n  {}\n  {}'.format(var_name, _var_calls[var_name], str(var).replace('\n', '\\n'), str(loaded_var).replace('\n', '\\n'))
            elif show=='max_diff':
                message = '{} @ {} does not match saved value.  max(|new-old|) = {}'.format(var_name, _var_calls[var_name], np.max(np.abs(var-loaded_var)))
            else:
                message = 'Match Failed'

            if raise_error:
                raise AssertionError(message)

            return False
        else:
            return True


def _get_random_finder_path(relative_path = ''):
    return get_artemis_data_path(os.path.join('random_finder', relative_path), make_local_dir=True)


def _get_tracker_file(var_name):
    return _get_random_finder_path('{}.pkl'.format(var_name))


def _compare(obj1, obj2):
    assert type(obj1) is type(obj2), "Types don't match: {}, {}".format(obj1, obj2)
    if isinstance(obj1, np.ndarray):
        assert np.array_equal(obj1, obj2), "Arrays not equal: {}, {}".format(obj1, obj2)
    else:
        assert obj1 == obj2

_var_readers = {}
_var_write_status = {}
_var_calls = {}


def _get_var_reader(var_name):

    with open(_get_tracker_file(var_name), 'rb') as f:
        while True:
            yield pickle.load(f)


def _get_next_saved_var_value(var_name):

    if var_name not in _var_readers:
        print('Found log of variable {}.  Comparing...'.format(var_name))
        _var_readers[var_name] = _get_var_reader(var_name)

    return next(_var_readers[var_name])


def _write_next_var_value(var, var_name):

    filename = _get_tracker_file(var_name)

    if not os.path.exists(filename):
        print('Creating log of variable {}'.format(var_name))

    with open(filename, 'ab') as f:
        pickle.dump(var, f)


def delete_vars(vars):
    path = _get_random_finder_path()
    if vars=='-all':
        vars = [os.path.splitext(var)[0] for var in os.listdir(path)]
    elif isinstance(vars, str):
        vars = [vars]
    else:
        assert isinstance(vars, (list, tuple))
    for var in vars:
        var_file = _get_tracker_file(var)
        if os.path.exists(var_file):
            os.remove(var_file)


def reset_variable_tracker():
    _var_readers.clear()
    _var_write_status.clear()
    _var_calls.clear()


if __name__ == '__main__':
    all_saved_vars = [os.path.splitext(var)[0] for var in os.listdir(_get_random_finder_path())]
    _vars = input('{}\nType <varname> to delete records for a varname, or -all to delete all. >>'.format('\n'.join(all_saved_vars) if len(all_saved_vars)>0 else '<No vars>'))
    delete_vars(_vars)



import SocketServer

from artemis.config import get_artemis_config_value
from pytest import raises

from artemis.remote.utils import get_local_ips, check_ssh_connection, check_if_port_is_free, get_socket

ip_addresses = [get_artemis_config_value(section="tests", option="remote_server",default_generator=lambda: "127.0.0.1")]
# ip_addresses=["127.0.0.1"]
if ip_addresses[0] not in get_local_ips():
    ip_addresses.append("127.0.0.1")

def test_check_ssh_connections():
    for ip_address in ip_addresses:
        if ip_address not in get_local_ips():
            check_ssh_connection(ip_address)


def test_check_if_port_is_free():
    for ip_address in ip_addresses:
        if ip_address not in get_local_ips():
            with raises(AssertionError):
                check_if_port_is_free(ip_address,80)
        else:
            sock, port = get_socket(ip_address,7005)
            # port is now not free:
            with raises(SocketServer.socket.error):
                check_if_port_is_free(ip_address, port)
            sock.close()
            # now no error to be expected
            check_if_port_is_free(ip_address,port)

if __name__ == "__main__":
    test_check_ssh_connections()
    test_check_if_port_is_free()



import itertools

__author__ = 'peter'


def cross_dict_dicts(*dicts):
    """
    Combine two or more dictionaries of dictionaries by turning every pairwise combination of their keys, and creating a
    new dict whose keys are tuples (containing these key-combinations) and whose values are the the combined dictionionaries.

    e.g.
        cross_dict_dicts({'a':{'aa': 1}, 'b':{'bb': 2}}, {'c': {'cc': 3}, 'd': {'dd': 4}})

        returns {
            ('a','c'):{'aa':1, 'cc':3},
            ('a','d'):{'aa':1, 'dd':4},
            ('b','c'):{'bb':2, 'cc':3},
            ('b','d'):{'bb':2, 'dd':4},
            }

    This can be useful if, for example, you want to try all several combinations of different arguments to a function.

    :param dicts: Dictionaries of dictionaries.
    :return: A Dictionary of dictionaries.
    """
    cross_dict = dict((keys, merge_dicts(*[d[k] for d, k in zip(dicts, keys)])) for keys in itertools.product(*[d.keys() for d in dicts]))
    return cross_dict


def merge_dicts(*dicts):
    """
    Given a collection of dictionaries, merge them.

    e.g.
        merge_dicts({'a': 1, 'b': 2}, {'c': 3, 'd': 4})
        returns {'a': 1, 'b': 2, 'c': 3, 'd': 4}

    Later dicts overwrite earlier ones.

    :param dicts: dictionaries.
    :return: A merged dictionary.
    """
    return dict((k, v) for d in dicts for k, v in d.items())



import atexit
import inspect
import logging
import os
import pickle
import shutil
import signal
import sys
import tempfile
import time
import traceback
from collections import OrderedDict
from contextlib import contextmanager
from datetime import datetime
from getpass import getuser
from uuid import getnode

from artemis.config import get_artemis_config_value
from artemis.fileman.local_dir import format_filename, make_file_dir, get_artemis_data_path, make_dir
from artemis.fileman.persistent_ordered_dict import PersistentOrderedDict
from artemis.general.display import CaptureStdOut
from artemis.general.functional import infer_function_and_derived_arg_values
from artemis.general.hashing import compute_fixed_hash
from artemis.general.should_be_builtins import nested
from artemis.general.test_mode import is_test_mode
from artemis.general.test_mode import set_test_mode

try:
    from enum import Enum
except ImportError:
    raise ImportError("Failed to import the enum package. This was added in python 3.4 but backported back to 2.4.  To install, run 'pip install --upgrade pip enum34'")

logging.basicConfig()
ARTEMIS_LOGGER = logging.getLogger('artemis')
ARTEMIS_LOGGER.setLevel(logging.INFO)

__author__ = 'peter'


class ExpInfoFields(Enum):
    NAME = 'Name'
    ID = 'Identifier'
    DIR = 'Directory'
    TIMESTAMP = 'Timestamp'
    ARGS = 'Args'
    FUNCTION = 'Function'
    MODULE = 'Module'
    FILE = 'File'
    STATUS = 'Status'
    N_FIGS = '# Figures Generated'
    FIGS = 'Figure Locs'
    RUNTIME = 'Run Time'
    VERSION = 'Version'
    NOTES = 'Notes'
    USER = 'User'
    MAC = 'MAC Address'
    PID = 'Process ID'


class ExpStatusOptions(Enum):
    STARTED = 'Running (or Killed)'
    ERROR = 'Error'
    STOPPED = 'Stopped by User'
    FINISHED = 'Ran Succesfully'
    CORRUPT = 'Corrupt'


ERROR_FLAG = object()


class ExperimentRecordInfo(object):

    def __init__(self, file_path, write_text_version=True):
        before, ext = os.path.splitext(file_path)
        assert ext == '.pkl', 'Your file-path must be a pickle'
        self._text_path = before + '.txt' if write_text_version else None
        self.persistent_obj = PersistentOrderedDict(file_path=file_path)

    def has_field(self, field):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        return field in self.persistent_obj

    def get_field(self, field, default = ERROR_FLAG):
        """
        :param field: A member of ExperimentRecordInfo.FIELDS
        :param default: Default to return if field does not exist (if left unspecified, we raise error)
        :return: The info for that field.
        """
        try:
            return self.persistent_obj[field]
        except KeyError:
            if default is ERROR_FLAG:
                raise
            else:
                return default

    def get_status_field(self):
        if self.has_field(ExpInfoFields.STATUS):
            return self.persistent_obj[ExpInfoFields.STATUS]
        else:
            return ExpStatusOptions.CORRUPT

    def set_field(self, field, value):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        if field == ExpInfoFields.STATUS:
            assert value in ExpStatusOptions, 'Status value must be in: {}'.format(ExpStatusOptions)
        with self.persistent_obj as pod:
            pod[field] = value
        if self._text_path is not None:
            with open(self._text_path, 'w') as f:
                f.write(self.get_text())

    def add_note(self, note):  # Currently unused
        if not self.has_field(ExpInfoFields.NOTES):
            self.set_field(ExpInfoFields.NOTES, [note])
        else:
            self.set_field(ExpInfoFields.NOTES, self.get_field(ExpInfoFields.NOTES) + [note])

    def get_notes(self):
        return [] if not self.has_field(ExpInfoFields.NOTES) else self.get_field(ExpInfoFields.NOTES)

    def get_text(self):
        if ExpInfoFields.VERSION not in self.persistent_obj:  # Old version... we must adapt
            return '\n'.join(
                '{}: {}'.format(key, self.get_field_text(key)) for key, value in self.persistent_obj.items())
        else:
            return '\n'.join(
                '{}: {}'.format(key.value, self.get_field_text(key)) for key, value in self.persistent_obj.items())

    def get_field_text(self, field, replacement_if_none=''):
        assert field in ExpInfoFields, 'Field must be a member of ExperimentRecordInfo.FIELDS'
        if not self.has_field(field):
            return replacement_if_none
        elif field is ExpInfoFields.STATUS:
            return self.get_field(field).value
        elif field is ExpInfoFields.ARGS:
            return ['{}={}'.format(k, v) for k, v in self.get_field(field)]
        else:
            return str(self.get_field(field))


class NoSavedResultError(Exception):
    def __init__(self, experiment_record_id):
        Exception.__init__(self, "Experiment Record {} has no saved result.".format(experiment_record_id))


class ExperimentRecord(object):
    """
    A Record of a run of an Experiment.  This object allows you to access data stored in the directory that was created
    for that run of the experiment.  Experiment Records are stored in ``~/artemis/experiments/``.
    """

    ERROR_FILE_NAME = 'errortrace.txt'

    def __init__(self, experiment_directory):
        self._experiment_directory = experiment_directory
        self._info = None

    @property
    def info(self):
        """
        :return: An ExperimentRecordInfo object, containing info about the experiment (name, runtime, etc)
        """
        if self._info is None:
            self._info = ExperimentRecordInfo(os.path.join(self._experiment_directory, 'info.pkl'))
        return self._info

    def __reduce__(self):
        return self.__class__, (self._experiment_directory, )

    def show_figures(self, hang=False):
        """
        Show all figures that were saved during the run of the experiment.
        :param hang: If True, and figures were saved matplotlib figures, hang execution until they are closed.
        """
        from artemis.plotting.saving_plots import show_saved_figure
        for i, loc in enumerate(self.get_figure_locs()):
            show_saved_figure(loc, title='{} Fig {}'.format(self.get_id(), i + 1))
        if hang and len(self.get_figure_locs())>0:
            from matplotlib import pyplot as plt
            plt.show()

    def get_log(self):
        """
        :return: The stdout generated during the run of this experiment.
        """
        log_file_path = os.path.join(self._experiment_directory, 'output.txt')
        assert os.path.exists(log_file_path), 'No output file found.  Maybe "%s" is not an experiment directory?' % (
        self._experiment_directory,)
        with open(log_file_path) as f:
            text = f.read()
        return text

    def list_files(self, full_path=False):
        """
        List files in experiment directory, relative to root.
        :param full_path: If true, list file with the full local path
        :return: A list of strings indicating the file paths.
        """
        paths = [os.path.join(root, filename) for root, _, files in os.walk(self._experiment_directory) for filename in
                 files]
        if not full_path:
            dir_length = len(self._experiment_directory)
            paths = [f[dir_length + 1:] for f in paths]
        return paths

    def open_file(self, filename, *args, **kwargs):
        """
        Open a file within the experiment record folder.
        Example Usage:

            with record.open_file('filename.txt') as f:
                txt = f.read()

        :param filename: Path within experiment directory (it can include subdirectories)
        :param args, kwargs: Forwarded to python's "open" function
        :return: A file object
        """
        full_path = os.path.join(self._experiment_directory, filename)
        make_file_dir(full_path)  # Make the path if it does not yet exist
        return open(full_path, *args, **kwargs)

    def get_figure_locs(self, include_directory=True):
        """
        Return a list of the paths of the figures saved in the experiment.
        :param include_directory: If True, return the full path.
        :return: A list of string file paths.
        """
        locs = [f for f in os.listdir(self._experiment_directory) if f.startswith('fig-')]
        if include_directory:
            return [os.path.join(self._experiment_directory, f) for f in locs]
        else:
            return locs

    def get_info_text(self):
        return self.info.get_text()

    def has_result(self):
        """
        :return: True if this record has a saved result.
        """
        return os.path.exists(os.path.join(self._experiment_directory, 'result.pkl'))

    def get_result(self, err_if_none = True):
        """
        Unpickle and return the "return value" of the experiment.
        :param err_if_none: If there is no saved return value, throw an exception if err_is_none, else just return None.
        :return: The return value from the experiment.
        """
        result_loc = os.path.join(self._experiment_directory, 'result.pkl')
        if os.path.exists(result_loc):
            with open(result_loc, 'rb') as f:
                result = pickle.load(f)
            return result
        elif err_if_none:
            raise NoSavedResultError(self.get_id())
        else:
            return None

    def save_result(self, result):
        file_path = get_local_experiment_path(os.path.join(self._experiment_directory, 'result.pkl'))
        make_file_dir(file_path)
        with open(file_path, 'wb') as f:
            pickle.dump(result, f, protocol=2)
            ARTEMIS_LOGGER.info('Saving Result for Experiment "{}"'.format(self.get_id(),))

    def get_id(self):
        """
        Get the id of this experiment record.  Generally in format '<datetime>-<experiment_name>'
        :return:
        """
        root, identifier = os.path.split(self._experiment_directory)
        return identifier

    def get_experiment(self):
        """
        Load the experiment associated with this record.
        Note that this will raise an ExperimentNotFoundError if the experiment has not been imported.
        :return: An Experiment object
        """
        from artemis.experiments.experiments import load_experiment
        return load_experiment(self.get_experiment_id())

    def get_experiment_id(self):
        return self.get_id()[27:]

    def get_timestamp(self):
        try:  # Faster, since we don't need to load the info object
            return time.mktime(datetime.strptime(self.get_id()[:26], '%Y.%m.%dT%H.%M.%S.%f').timetuple())
        except:
            return time.mktime(datetime.strptime(self.info.get_field(ExpInfoFields.TIMESTAMP), '%Y-%m-%d %H:%M:%S.%f').timetuple())

    def get_dir(self):
        """
        :return: The directory associated with this experiment record.
        """
        return self._experiment_directory

    def get_args(self):
        """
        Get the arguments with which this record was run.
        :return: An OrderedDict((arg_name -> arg_value))
        """
        return OrderedDict(self.info.get_field(ExpInfoFields.ARGS))

    def get_status(self):
        try:
            return self.info.get_field(ExpInfoFields.STATUS)
        except KeyError:
            return ExpStatusOptions.CORRUPT

    def load_figures(self):
        """
        :return: A list of matplotlib figures generated in the experiment.  The figures will not be drawn yet, so you
            will have to call plt.show() to draw them or plt.draw() to draw them.
        """
        locs = self.get_figure_locs()
        figs = []
        for fig_path in locs:
            assert fig_path.endswith('.pkl'), 'Figure {} was not saved as a pickle, so it cannot be reloaded.'.format(fig_path)
            with open(fig_path, 'rb') as f:
                figs.append(pickle.load(f))
        return figs

    def delete(self):
        """
        Delete this experiment record from disk.
        """
        shutil.rmtree(self._experiment_directory)

    def args_valid(self, last_run_args=None, current_args=None):
        """
        :return: True if the experiment arguments have not changed
            False if they have changed
            None if it cannot be determined because arguments are not hashable objects.
        """
        if last_run_args is None:  # Cast to dict (from OrderedDict) because different arg order shouldn't matter
            last_run_args = self.info.get_field(ExpInfoFields.ARGS)  # A list of 2-tuples
        if current_args is None:
            current_args = dict(self.get_experiment().get_args())
        try:
            return compute_fixed_hash(dict(last_run_args), try_objects=True) == compute_fixed_hash(dict(current_args), try_objects=True)
        except NotImplementedError:  # Happens when we have unhashable arguments
            return None

    def get_error_trace(self):
        """
        Get the error trace, or return None if there is no error trace.
        :return:
        """
        file_path = os.path.join(self._experiment_directory, self.ERROR_FILE_NAME)
        if os.path.exists(file_path):
            with open(file_path) as f:
                return f.read()
        else:
            return None

    def write_error_trace(self, print_too = True):
        file_path = os.path.join(self._experiment_directory, self.ERROR_FILE_NAME)
        assert not os.path.exists(file_path), 'Error trace has already been created in this experiment... Something fishy going on here.'
        with open(file_path, 'w') as f:
            error_text = traceback.format_exc()
            f.write(error_text)
        if print_too:
            print(error_text)

    def kill(self, assert_alive = True):
        status = self.info.get_field(ExpInfoFields.STATUS)
        if status is ExpStatusOptions.STARTED:
            pid = self.info.get_field(ExpInfoFields.PID)
            try:
                os.kill(pid, signal.SIGKILL)
            except OSError:
                print('Process {} appears to be already dead.  '.format(pid))
            self.info.set_field(ExpInfoFields.STATUS, ExpStatusOptions.STOPPED)
        elif assert_alive:
            raise Exception('Cannot kill a process with status "{}", for it is already dead.'.format(status))


_CURRENT_EXPERIMENT_RECORD = None


@contextmanager
def hold_current_experiment_record(experiment_record):
    global _CURRENT_EXPERIMENT_RECORD
    assert _CURRENT_EXPERIMENT_RECORD is None, "It seems that you are trying to start an experiment withinin an experiment.  This is not allowed!"
    _CURRENT_EXPERIMENT_RECORD = experiment_record
    try:
        yield
    except Exception as err:
        raise err
    finally:
        _CURRENT_EXPERIMENT_RECORD = None


def is_matplotlib_imported():
    return 'matplotlib' in sys.modules


@contextmanager
def record_experiment(identifier='%T-%N', name='unnamed', print_to_console=True, show_figs=None,
                      save_figs=True, saved_figure_ext='.fig.pkl', use_temp_dir=False, date=None, prefix=None):
    """
    :param identifier: The string that uniquely identifies this experiment record.  Convention is that it should be in
        the format
    :param name: Base-name of the experiment
    :param print_to_console: If True, print statements still go to console - if False, they're just rerouted to file.
    :param show_figs: Show figures when the experiment produces them.  Can be:
        'hang': Show and hang
        'draw': Show but keep on going
        False: Don't show figures
    """
    # Note: matplotlib imports are internal in order to avoid trouble for people who may import this module without having
    # a working matplotlib (which can occasionally be tricky to install).
    if date is None:
        date = datetime.now()
    identifier = format_filename(file_string=identifier, base_name=name, current_time=date)

    if show_figs is None:
        show_figs = 'draw' if is_test_mode() else 'hang'

    assert show_figs in ('hang', 'draw', False)

    if use_temp_dir:
        experiment_directory = tempfile.mkdtemp()
        atexit.register(lambda: shutil.rmtree(experiment_directory))
    else:
        experiment_directory = get_local_experiment_path(identifier)

    make_dir(experiment_directory)
    this_record = ExperimentRecord(experiment_directory)

    # Create context that sets the current experiment record
    # and the context which captures stdout (print statements) and logs them.
    contexts = [
        hold_current_experiment_record(this_record),
        CaptureStdOut(log_file_path=os.path.join(experiment_directory, 'output.txt'), print_to_console=print_to_console, prefix=prefix)
        ]

    if is_matplotlib_imported():
        from artemis.plotting.manage_plotting import WhatToDoOnShow
        # Add context that modifies how matplotlib figures are shown.
        contexts.append(WhatToDoOnShow(show_figs))
        if save_figs:
            from artemis.plotting.saving_plots import SaveFiguresOnShow
            # Add context that saves figures when show is called.
            contexts.append(SaveFiguresOnShow(path=os.path.join(experiment_directory, 'fig-%T-%L' + saved_figure_ext)))

    with nested(*contexts):
        yield this_record


def is_in_experiment():
    """
    :return: True if this function is called from within a running experiment, False otherwise.
    """
    return _CURRENT_EXPERIMENT_RECORD is not None


def get_current_experiment_record():
    if _CURRENT_EXPERIMENT_RECORD is None:
        raise Exception("No experiment is currently running!")
    return _CURRENT_EXPERIMENT_RECORD


def get_current_experiment_id():
    """
    :return: A string identifying the current experiment
    """
    return get_current_experiment_record().get_experiment_id()


def get_current_record_id():
    """
    :return: A string identifying the current experiment
    """
    return get_current_experiment_record().get_id()


def get_current_record_dir(default_if_none = True):
    """
    The directory in which the results of the current experiment are recorded.
    """
    if _CURRENT_EXPERIMENT_RECORD is None and default_if_none:
        return get_artemis_data_path('experiments/default/', make_local_dir=True)
    else:
        return get_current_experiment_record().get_dir()


def open_in_record_dir(filename, *args, **kwargs):
    """
    Open a file in the given experiment directory.  Usage:

    with open_in_record_dir('myfile.txt', 'w') as f:
        f.write('blahblahblah')

    :param filename: The name of the file, relative to your experiment directory,
    :param args,kwargs: See python built-in "open" function
    :yield: The file object
    """
    return get_current_experiment_record().open_file(filename, *args, **kwargs)


def record_id_to_experiment_id(record_id):
    return load_experiment_record(record_id).get_experiment_id()


def delete_experiment_with_id(experiment_identifier):
    if experiment_exists(experiment_identifier):
        load_experiment_record(experiment_identifier).delete()


def get_experiment_dir():
    path = os.path.expanduser(get_artemis_config_value(section="experiments", option="experiment_directory", write_default=True, default_generator=lambda: get_artemis_data_path('experiments')))
    if not os.path.exists(path):
        make_dir(path)
    return path


def get_local_experiment_path(identifier):
    return os.path.join(get_experiment_dir(), identifier)


def experiment_exists(identifier):
    local_path = get_local_experiment_path(identifier)
    return os.path.exists(local_path)


def merge_experiment_dicts(*dicts):
    """
    Merge dictionaries of experiments, checking that names are unique.
    """
    merge_dict = OrderedDict()
    for d in dicts:
        assert not any(k in merge_dict for k in d), "Experiments %s has been defined twice." % (
        [k for k in d.keys() if k in merge_dict],)
        merge_dict.update(d)
    return merge_dict


def filter_experiment_ids(record_ids, expr=None, experiment_ids=None):
    if expr is not None:
        record_ids = [e for e in record_ids if expr in e]
    if experiment_ids is not None:
        record_ids = [record_id for record_id in record_ids if ExperimentRecord(record_id).get_experiment_id() in experiment_ids]
    return record_ids


def get_all_record_ids(experiment_ids=None, filters=None):
    """
    :param experiment_ids: A list of experiment names
    :param filters: A list or regular expressions for matching experiments.
    :return: A list of experiment identifiers.
    """
    expdir = get_experiment_dir()
    ids = [e for e in os.listdir(expdir) if os.path.isdir(os.path.join(expdir, e))]
    ids = filter_experiment_ids(record_ids=ids, experiment_ids=experiment_ids)
    if filters is not None:
        for expr in filters:
            ids = filter_experiment_ids(record_ids=ids, expr=expr)
    ids = sorted(ids)
    return ids


def experiment_id_to_record_ids(experiment_identifier):
    """
    :param experiment_identifier: The name of the experiment
    :param filter_status: An ExpStatusOptions enum, indicating that you only want experiments with this status.
        e.g. ExpStatusOptions.FINISHED
    :return: A list of records for this experiment, temporal order
    """
    matching_records = get_all_record_ids(experiment_ids=[experiment_identifier])
    return sorted(matching_records)


def has_experiment_record(experiment_identifier):
    return len(experiment_id_to_record_ids(experiment_identifier)) != 0


def load_experiment_record(record_id):
    """
    Load an ExperimentRecord based on the identifier
    :param record_id: A string identifying the experiment record
    :return: An ExperimentRecord object
    """
    path = os.path.join(get_experiment_dir(), record_id)
    return ExperimentRecord(path)


def clear_experiment_records(ids):
    """
    Delete all experiments with ids in the list, or all experiments if ids is None.
    :param ids: A list of experiment ids, or None to remove all.
    """
    folder = get_experiment_dir()
    for exp_id in ids:
        exp_path = os.path.join(folder, exp_id)
        ExperimentRecord(exp_path).delete()


def save_figure_in_record(name, fig=None, default_ext='.pkl'):
    '''
    Saves the given figure in the experiment directory. If no figure is passed, plt.gcf() is saved instead.
    :param name: The name of the figure to be saved
    :param fig: The figure to be saved, can be None
    :param default_ext: See artemis.plotting.saving_plots.save_figure() for information
    :return: The path to the figure
    '''
    import matplotlib.pyplot as plt
    from artemis.plotting.saving_plots import save_figure
    if fig is None:
        fig = plt.gcf()
    save_path = os.path.join(get_current_record_dir(), name)
    save_figure(fig, path=save_path, default_ext=default_ext)
    return save_path


def run_and_record(function, experiment_id, print_to_console=True, show_figs=None, test_mode=None, keep_record=None,
        raise_exceptions=True, notes = (), prefix=None, **experiment_record_kwargs):
    """
    Run an experiment function.  Save the console output, return values, and any matplotlib figures generated to a new
    experiment folder in ~/.artemis/experiments

    :param function: A function which takes no args.
    :param experiment_id: The name under which you'd like to save this run of this experiment.
    :param print_to_console: Show the print output in the console (as well as saving it)
    :param show_figs:
    :param test_mode:
    :param keep_record:
    :param raise_exceptions:
    :param notes:
    :param experiment_record_kwargs:
    :return: The ExperimentRecord object
    """

    if test_mode is None:
        test_mode = is_test_mode()

    old_test_mode = is_test_mode()
    set_test_mode(test_mode)
    ARTEMIS_LOGGER.info('{border} {mode} Experiment: {name} {border}'
        .format(border='=' * 10, mode="Testing" if test_mode else "Running", name=experiment_id))
    EIF = ExpInfoFields
    date = datetime.now()
    with record_experiment(name=experiment_id, print_to_console=print_to_console, show_figs=show_figs,
            use_temp_dir=not keep_record, date=date, prefix=prefix, **experiment_record_kwargs) as exp_rec:
        start_time = time.time()
        try:

            exp_rec.info.set_field(ExpInfoFields.NAME, experiment_id)
            exp_rec.info.set_field(ExpInfoFields.ID, exp_rec.get_id())
            exp_rec.info.set_field(ExpInfoFields.DIR, exp_rec.get_dir())
            root_function, args = infer_function_and_derived_arg_values(function)
            exp_rec.info.set_field(EIF.ARGS, list(args.items()))
            # root_function = self.get_root_function()
            exp_rec.info.set_field(EIF.FUNCTION, root_function.__name__)
            exp_rec.info.set_field(EIF.TIMESTAMP, date)
            module = inspect.getmodule(root_function)
            exp_rec.info.set_field(EIF.MODULE, module.__name__)
            exp_rec.info.set_field(EIF.FILE, module.__file__ if hasattr(module, '__file__') else '<unknown>')
            exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.STARTED)
            exp_rec.info.set_field(EIF.USER, getuser())
            exp_rec.info.set_field(EIF.MAC, ':'.join(("%012X" % getnode())[i:i+2] for i in range(0, 12, 2)))
            exp_rec.info.set_field(EIF.PID, os.getpid())
            if inspect.isgeneratorfunction(root_function):
                for result in function():
                    exp_rec.save_result(result)
            else:
                result = function()
                exp_rec.save_result(result)
            exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.FINISHED)
        except KeyboardInterrupt:
            exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.STOPPED)
            exp_rec.write_error_trace(print_too=False)
            raise
        except Exception:
            exp_rec.info.set_field(EIF.STATUS, ExpStatusOptions.ERROR)
            exp_rec.write_error_trace(print_too=not raise_exceptions)
            if raise_exceptions:
                raise
            else:
                return exp_rec
        finally:
            exp_rec.info.set_field(EIF.RUNTIME, time.time() - start_time)
            fig_locs = exp_rec.get_figure_locs(include_directory=False)
            exp_rec.info.set_field(EIF.N_FIGS, len(fig_locs))
            exp_rec.info.set_field(EIF.FIGS, fig_locs)

    for n in notes:
        exp_rec.info.add_note(n)

    ARTEMIS_LOGGER.info('{border} Done {mode} Experiment: {name} {border}'.format(border='=' * 10, mode="Testing" if test_mode else "Running", name=experiment_id))
    set_test_mode(old_test_mode)

    return exp_rec



from collections import OrderedDict

import pytest
from pytest import raises

from artemis.general.duck import Duck, InvalidKeyError, InitializerTooShallowError
from artemis.general.hashing import compute_fixed_hash
import numpy as np

from artemis.general.should_be_builtins import izip_equal


def _get_standard_test_duck():
    a = Duck()
    a['a', 'aa1'] = 1
    a['a', 'aa2'] = 2
    a['b', 0, 'subfield1'] = 4
    a['b', 0, 'subfield2'] = 5
    a['b', 1, 'subfield1'] = 6
    a['b', 1, 'subfield2'] = 7
    return a


def test_so_demo():

    # Demo 1: Dynamic assignment
    a = Duck()
    a['a', 'aa1'] = 1
    a['a', 'aa2'] = 2
    a['b', 0, 'subfield1'] = 4
    a['b', 0, 'subfield2'] = 5
    a['b', 1, 'subfield1'] = 6
    a['b', 1, 'subfield2'] = 7

    assert list(a['b', 1, :]) == [6, 7]
    assert a['b', :, 'subfield1'] == [4, 6]
    assert a['a', 'aa2'] == 2
    assert np.array_equal(a['b'].to_array(), [[4, 5], [6, 7]])
    with pytest.raises(KeyError):  # This should raise an error because key 'a' does not have subkeys 1, 'subfield1'
        x = a[:, 1, 'subfield1']

    # Demo 2: Sequential and Sliced Assignment
    # Here we show another way to create the same structure as above
    # 1) You can assign to a slice
    # 2) You can use the "next" builtin like an index to append to the structure.
    b = Duck()
    b['a', :] = {'aa1': 1, 'aa2': 2}  # Note: when assigning with dict, keys are sorted before insertion (OrderedDict order is kept though).
    b['b', next, :] = {'subfield1': 4, 'subfield2': 5}
    b['b', next, :] = {'subfield1': 6, 'subfield2': 7}
    assert b==a


def test_dict_assignment():

    a = Duck()  # When assigning with a dict we first sort keys.  Here we just verify that keys remain sorted
    a[next, :] = {letter: number for letter, number in izip_equal('abcdefghijklmnopqrstuvwxyz', range(1, 27))}
    a[next, :] = {letter: number for letter, number in izip_equal('abcdefghijklmnopqrstuvwxyz', range(27, 27+26))}
    assert list(a[0].keys()) == [char for char in 'abcdefghijklmnopqrstuvwxyz']
    assert list(a[0].values()) == list(range(1, 27))
    assert list(a[1].keys()) == [char for char in 'abcdefghijklmnopqrstuvwxyz']
    assert list(a[1].values()) == list(range(27, 27+26))


def test_reasonable_errors_on_wrong_keys():

    a = Duck()
    a['a']=1
    with pytest.raises(InvalidKeyError):
        a[1] = 'a'
    with pytest.raises(InvalidKeyError):
        a[next]='a'

    a = Duck()
    a[next] = 'a'
    with pytest.raises(InvalidKeyError):
        a['a'] = 1
    with pytest.raises(InvalidKeyError):
        a[10] = 'a'  # Too high!


def test_string_slices():

    a = Duck()
    for i, char in enumerate('abcdefghijklmnopqrstuvwxyz'[::-1]):  # Note the reverse
        a[char]=26-i
    assert list(a['q':'m'].items()) == [('q', 17), ('p', 16), ('o', 15), ('n', 14)]
    assert list(a['m':'q'].items()) == []
    assert list(a['m':'q':-1].items()) == [('m', 13), ('n', 14), ('o', 15), ('p', 16)]


def test_dictarraylist():

    # We build the same thing in four different ways

    a = Duck()
    a['a', 'aa1'] = 1
    a['a', 'aa2'] = 2
    a['b', 0, 'subfield1'] = 4
    a['b', 0, 'subfield2'] = 5
    a['b', 1, 'subfield1'] = 6
    a['b', 1, 'subfield2'] = 7
    a['b', 2, 'subfield2'] = 9  # Note: Out of order assignment here!
    a['b', 2, 'subfield1'] = 8

    b = Duck.from_struct(a.to_struct())
    assert a.to_struct() == b.to_struct()

    c = Duck()
    c['a', :] = OrderedDict([('aa1', 1), ('aa2', 2)])
    c['b', next, :] = OrderedDict([('subfield1', 4), ('subfield2', 5)])
    c['b', next, :] = OrderedDict([('subfield1', 6), ('subfield2', 7)])
    c['b', next, :] = OrderedDict([('subfield2', 9), ('subfield1', 8)])

    d = Duck()
    d['a', :] = OrderedDict([('aa1', 1), ('aa2', 2)])
    d['b', :, :] = [OrderedDict([('subfield1', 4), ('subfield2', 5)]), OrderedDict([('subfield1', 6), ('subfield2', 7)]), OrderedDict([('subfield2', 9), ('subfield1', 8)])]

    e = Duck()
    e['a', ...] = OrderedDict([('aa1', 1), ('aa2', 2)])
    e['b', ...] = [OrderedDict([('subfield1', 4), ('subfield2', 5)]), OrderedDict([('subfield1', 6), ('subfield2', 7)]), OrderedDict([('subfield2', 9), ('subfield1', 8)])]

    f = a[:]

    g = a[:, :]

    for i, same_struct in enumerate([a, b, c, d, e, f, g]):

        print ('Test {}'.format(i))
        assert a.to_struct()==same_struct.to_struct()
        assert a==same_struct
        assert compute_fixed_hash(a.to_struct()) == compute_fixed_hash(b.to_struct())

        assert same_struct['a', 'aa2'] == 2
        assert list(same_struct['b', 1, :].values()) == [6, 7]
        assert list(same_struct['b', :, 'subfield1'].values()) == [4, 6, 8]
        assert same_struct['b', :, :].deepvalues() == [[4, 5], [6, 7], [9, 8]]  # Note that the order-switching gets through.
        assert same_struct['b', :, ['subfield1', 'subfield2']].deepvalues() == [[4, 5], [6, 7], [8, 9]]
        assert same_struct['b', :, :].to_struct() == [OrderedDict([('subfield1', 4), ('subfield2', 5)]), OrderedDict([('subfield1', 6), ('subfield2', 7)]), OrderedDict([('subfield2', 9), ('subfield1', 8)])]
        assert same_struct['b', 1:, :].to_struct() == [OrderedDict([('subfield1', 6), ('subfield2', 7)]), OrderedDict([('subfield2', 9), ('subfield1', 8)])]

        with pytest.raises(KeyError):
            x = same_struct['a', 'aa3']

        with pytest.raises(KeyError):
            x = same_struct['a', 1, 'subfield1']

        with pytest.raises(IndexError):
            x = same_struct['b', 4, 'subfield1']

        with pytest.raises(KeyError):  # This should raise an error because key 'a' does not have subkeys 1, 'subfield1'
            x = same_struct[:, 1, 'subfield1']

        assert np.array_equal(same_struct['b'].to_array(), [[4, 5], [6, 7], [8, 9]])  # Note... the order is corrected.
        with pytest.raises(InvalidKeyError):
            assert np.array_equal(same_struct.to_array())  # Note... the order is corrected.

        new_struct = same_struct.arrayify_axis(axis=1, subkeys=('b',))
        assert np.array_equal(new_struct['b', 'subfield1'], [4, 6, 8])
        assert np.array_equal(new_struct['b', 'subfield2'], [5, 7, 9])


def test_simple_growing():
    a = Duck()
    for i in range(10):
        a[next] = i*2
    assert np.array_equal(a.to_array(), np.arange(0, 20, 2))


def test_open_key():

    a = Duck()
    a['a', :] = [3, 4]
    with pytest.raises(KeyError):
        a['b']
    c = a.open('b')
    c[next] = 5
    c[next] = 6
    assert a['b', 0]==5
    assert a['b', 1]==6


def test_open_next():

    a = Duck()
    a['a'] = Duck()
    aa = a['a'].open(next)
    aa['c'] = 4
    aa['4'] = 5
    aa = a['a'].open(next)
    aa['c'] = 6
    aa['4'] = 7


def test_to_struct():

    a = _get_standard_test_duck()

    b = a.to_struct()
    assert b==OrderedDict([
        ('a', OrderedDict([('aa1', 1), ('aa2', 2)])),
        ('b', [
            OrderedDict([('subfield1', 4), ('subfield2', 5)]),
            OrderedDict([('subfield1', 6), ('subfield2', 7)])
            ])])


def test_next_elipsis_assignment():

    a = Duck()
    a['a', next, ...] = OrderedDict([('b', [1, 2])])
    a['a', next, ...] = OrderedDict([('b', [3, 4])])
    # Should lead to form:
    # 'a'
    #   0
    #     'b'
    #       0: 1
    #       1: 2
    #   1
    #     'b'
    #       0: 3
    #       1: 4

    # If we arrayify axis, 1, we expect struct:

    # Should lead to form:
    # 'a'
    #   'b'
    #     0: [1, 3]
    #     1: [2, 4]
    assert np.array_equal(a.arrayify_axis(axis=1, subkeys='a')['a', 'b'], [[1, 3], [2, 4]])

    a = Duck()
    a['a', next, ...] = OrderedDict([('b', np.array([1, 2]))])
    a['a', next, ...] = OrderedDict([('b', np.array([3, 4]))])

    # Now, a is no longer broken into ...  So we have to be careful with these elipses!
    # 'a'
    #   0
    #     'b': [1, 2]
    #   1
    #     'b': [3, 4]

    # If we arrayify axis, 1, we expect struct:

    # Should lead to form:
    # 'a'
    #   'b'
    #     0: [1, 2]
    #     1: [3, 4]
    assert np.array_equal(a.arrayify_axis(axis=1, subkeys='a')['a', 'b'], [[1, 2], [3, 4]])


def test_slice_assignment():

    deep_struct = {'a': [1,{'c': 2, 'd': 3}], 'b': [1, 2]}

    a = Duck()
    a[:] = deep_struct
    assert list(a.items(depth='full')) == [
        (('a', ), [1,{'c': 2, 'd': 3}]),
        (('b', ), [1, 2])
        ]
    a = Duck()
    a[:, :] = deep_struct
    assert list(a.items(depth='full')) == [
        (('a', 0), 1),
        (('a', 1), {'c': 2, 'd': 3}),
        (('b', 0), 1),
        (('b', 1), 2)
        ]
    a = Duck()
    with raises(InitializerTooShallowError):
        a[:, :, :] = deep_struct

    a = Duck()
    a[...] = deep_struct
    assert list(a.items(depth='full')) == [
        (('a', 0), 1),
        (('a', 1, 'c'), 2),
        (('a', 1, 'd'), 3),
        (('b', 0), 1),
        (('b', 1), 2)
        ]

    a = Duck()
    with raises(InvalidKeyError):  # It would be nice to add this functionality, but maybe later.
        a[..., next] = deep_struct
        # This commant would mean "turn every leaf node in deep_struct into a list".  It has the nice property
        # of automatically asserting that deep_struct maintains the same structure between iterations.


def test_arrayify_empty_stuct():

    a = Duck.from_struct(OrderedDict([('a', [])]))
    b = a.arrayify_axis(axis=1, subkeys='a')
    assert isinstance(b['a'], np.ndarray)
    assert np.array_equal(b['a'], [])

    a = Duck.from_struct(OrderedDict([('a', [1, 2])]))
    b = a.arrayify_axis(axis=1, subkeys='a')
    assert isinstance(b['a'], np.ndarray)
    assert np.array_equal(b['a'], [1, 2])


def test_slice_on_start():

    a = Duck()
    a['testing'] = Duck()
    b = a.arrayify_axis(axis=1, subkeys='testing')
    assert np.array_equal(b['testing'], [])


def test_assign_tuple_keys():

    a = Duck()
    a['training'] = Duck()
    a['testing', next]=4
    b = a.arrayify_axis(axis=1, subkeys='testing')
    b = b.arrayify_axis(axis=1, subkeys='training', inplace=True).to_struct()
    a['training', next, ...] = OrderedDict([('ops', OrderedDict([(('a', 'b'), 1), (('a', 'c'), 2)]))])
    a['training', next, ...] = OrderedDict([('ops', OrderedDict([(('a', 'b'), 3), (('a', 'c'), 4)]))])
    a.arrayify_axis(axis=1, subkeys='training').to_struct()
    a['training', next, ...] = OrderedDict([('ops', OrderedDict([(('a', 'b'), 1), (('a', 'c'), 2)]))])
    pass


def test_broadcast_bug():

    a = _get_standard_test_duck()

    assert isinstance(a['b', :], Duck)
    assert np.array_equal(a['b'].to_array(), [[4, 5], [6, 7]])
    assert a['b']==a['b', :, :]
    assert np.array_equal(a['b', :, :].to_array(), [[4, 5], [6, 7]])  # This failed, but now is fixed.


def test_key_values():

    a = _get_standard_test_duck()

    assert list(a.keys()) == ['a', 'b']
    assert list(a.values()) == [a['a'], a['b']]
    assert list(a.keys(depth=1)) == [('a', ), ('b', )]
    assert list(a.values(depth=1)) == [a['a'], a['b']]
    assert list(a.keys(depth=2)) == [('a', 'aa1'), ('a', 'aa2'), ('b', 0), ('b', 1)]
    assert list(a.values(depth=2)) == [1, 2, a['b', 0], a['b', 1]]
    assert list(a.keys(depth='full')) == [('a', 'aa1'), ('a', 'aa2'), ('b', 0, 'subfield1'), ('b', 0, 'subfield2'), ('b', 1, 'subfield1'), ('b', 1, 'subfield2')]
    assert list(a.values(depth='full')) == [1, 2, 4, 5, 6, 7]
    assert list(a.items(depth='full')) == list(zip(a.keys(depth='full'), a.values(depth='full')))


_expected_description = """<Duck with 2 keys: ['a', 'b']>
| a: 
| | aa1: 1
| | aa2: 2
| b: 
| | 0: 
| | | subfield1: 4
| | | subfield2: 5
| | 1: 
| | | subfield1: 6
| | | subfield2: 7"""


_extended_expected_description = """<Duck with 2 keys: ['a', 'b']>
| a: 
| | aa1: 1
| | aa2: 2
| b: 
| | 0: 
| | | subfield1: 4
| | | subfield2: 5
| | 1: 
| | | subfield1: 6
| | | subfield2: 7
| | 2: 
| | | subfield1: 5
| | 3: 
| | | subfield1: 5
| | 4: 
| | | subfield1: 5
| | (... Omitting 2 of 6 elements ...)"""



def test_description():
    a = _get_standard_test_duck()
    assert a.description()==_expected_description

    a['b', 2, 'subfield1']=5
    a['b', 3, 'subfield1']=5
    a['b', 4, 'subfield1']=5
    a['b', 5, 'subfield1']=5

    assert a.description()==_extended_expected_description


def test_duck_array_build():

    a = Duck()
    count = 0
    for i in range(3):
        for j in range(4):
            a[i, j] = count
            count+=1
    desired_array = np.arange(12).reshape(3, 4)
    assert a==desired_array
    assert np.array_equal(a.to_array(), desired_array)


def test_split_get_assign():

    a = Duck()
    c=0
    for i in ['a', 'b']:
        for j in [0, 1]:
            for k in ['c', 'd']:
                a[i, j, k] = c
                c+=1
    b = a[:, 1, :]
    print (a[:, 1, :].description())
    assert list(b.items(depth='full')) == [(('a', 'c'), 2), (('a', 'd'), 3), (('b', 'c'), 6), (('b', 'd'), 7)]

    with pytest.raises(InvalidKeyError):
        a[:, 1, :] = b


def test_assign_from_struct():

    a = Duck()
    a[:] = [2, 3, 4, 5]
    assert list(a.items()) == [(0, 2), (1, 3), (2, 4), (3, 5)]
    a[:] = dict(a=1, b=2, c=3)
    assert list(a.items()) == [('a', 1), ('b', 2), ('c', 3)]


def test_arrayify_axis_demo():

    a = Duck()
    a[0, 'x'] = 1
    a[0, 'y'] = 2
    a[1, 'x'] = 3
    a[1, 'y'] = 4
    b = a.arrayify_axis(axis=0)
    assert np.array_equal(b['x'], [1, 3])
    assert np.array_equal(b['y'], [2, 4])


if __name__ == '__main__':
    test_so_demo()
    test_dict_assignment()
    test_dictarraylist()
    test_simple_growing()
    test_open_key()
    test_open_next()
    test_to_struct()
    test_next_elipsis_assignment()
    test_slice_assignment()
    test_arrayify_empty_stuct()
    test_slice_on_start()
    test_assign_tuple_keys()
    test_broadcast_bug()
    test_key_values()
    test_description()
    test_duck_array_build()
    test_split_get_assign()
    test_assign_from_struct()
    test_arrayify_axis_demo()
    test_string_slices()
    test_reasonable_errors_on_wrong_keys()



'''
This file is required to perform plotting server related tests. Its purpose is to be executed in a seperate process or on a remote machine, automatically.
'''

from __future__ import print_function
from argparse import ArgumentParser
import sys
import atexit
import time

def atexit_function():
    print("atexit called")
    sys.exit(0)

def sleep_function():
    atexit.register(lambda:print("Interrupted"))
    time.sleep(100)

def hanging_sleep_function():
    atexit.register(lambda:time.sleep(100))
    time.sleep(100)

def sleep_N(N):
    time.sleep(N)

def remote_graphics():
    from matplotlib import pyplot as plt
    plt.figure()
    # plt.show()
    plt.draw()
    plt.pause(0.1)
    print("Success")

def count(N):
    for i in range(N):
        print(i)
        time.sleep(1.0)

def iter_print():
    for i in range(10):
        if i%2==0:
            sys.stdout.write(str(i))
            sys.stdout.write("\n")
            # sys.stdout.flush()
        else:
            sys.stderr.write(str(i))
            sys.stderr.write("\n")
            # sys.stderr.flush()
        time.sleep(1.0)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument('--callback',type=str,default="") # to know which test is calling this script and to setup the behaviour accordingly.
    args = parser.parse_args()
    fun = {
        "success_function":lambda: print("Success"),
        "sleep_function":sleep_function,
        "hanging_sleep_function":hanging_sleep_function,
        "remote_graphics":remote_graphics,
        "count_low":lambda: count(5),
        "count_high":lambda: count(50),
        "short_sleep":lambda: sleep_N(10),
        "iter_print":iter_print,
        }[args.callback]
    fun()




from __future__ import print_function

from six.moves import queue
import atexit
import base64
import inspect
import shlex
import signal
import subprocess
import sys
import threading
import time
import uuid

import os
import pickle
import pipes


from six import string_types

from artemis.general.should_be_builtins import file_path_to_absolute_module
from artemis.remote import remote_function_run_script
from artemis.remote.plotting.utils import handle_socket_accepts

from artemis.config import get_artemis_config_value
from artemis.remote.utils import get_local_ips, get_socket, get_ssh_connection, check_pid


class ChildProcess(object):
    '''
    Generic Child Process
    '''
    counter=1
    def __init__(self, command, ip_address = 'localhost', name=None, take_care_of_deconstruct=False, set_up_port_for_structured_back_communication=False):
        '''
        Creates a ChildProcess
        :param ip_address: The command will be executed at this ip_address
        :param command: the command to execute.
        :param name: optional name. If not set, will be process_i, with i a global counter
        :param take_care_of_deconstruct: If set to True, deconstruct() is registered at exit
        :param port_for_structured_back_communication: Needs to be implemented according to the properties of the child process (see PythonChildProcess)
        :return:
        '''
        if name is None:
            name = "process_%s"%ChildProcess.counter
        if ip_address=='localhost':
            ip_address='127.0.0.1'
        ChildProcess.counter += 1
        self.name = name
        self.ip_address = ip_address
        self.local_process = self.ip_address in get_local_ips()
        self.set_up_port_for_structured_back_communication = set_up_port_for_structured_back_communication
        self.queue_from_cp = None

        self.command = command
        self.id = uuid.uuid4()
        self.channel = None
        self.cp_started = False
        self.take_care_of_deconstruct = take_care_of_deconstruct
        if self.take_care_of_deconstruct:
            atexit.register(self.deconstruct)

    def prepare_command(self,command):
        '''
        All the stuff that I need to prepare for the command to definitely work
        :param command:
        :return:
        '''
        if self.is_local():
            return command
        else:
            if isinstance(command,list):
                command = " ".join(pipes.quote(c) for c in command)

            return self.get_extended_command(command)
        return command

    def get_extended_command(self,command):
        return "echo $$ ; exec %s"%command

    def deconstruct(self, signum=signal.SIGKILL, system_signal=False):
        '''
        This completely and safely deconstructs a remote connection. It might also be called at program shutdown, if take_care_of_deconstruct is set to True
        kills itself if alive, then closes remote connection if applicable
        :return:
        '''
        if self.cp_started:
            if signum == signal.SIGKILL:
                self.kill(signum=signum)
            elif signum == signal.SIGINT:
                self.kill(signum=signum)
                counter = 0
                while counter < 3:
                    if self.is_alive():
                        # 3 seconds grace period
                        counter +=1
                        time.sleep(1.0)
                    else:
                        break
                if self.is_alive():
                    # grace perior exceeded, terminating
                    self.kill(signum=signal.SIGTERM)


    def is_local(self):
        return self.local_process

    def get_name(self):
        return self.name

    def get_ip(self):
        return self.ip_address

    def get_id(self):
        return self.ip_address + "_" + str(self.id)

    def _assign_pid(self,pid):
        self.pid = pid

    def get_pid(self):
        assert self.cp_started, "This ChildProcess has not been executed yet, no pid available yet. run execute_child_process() first"
        return self.pid

    def get_ssh_connection(self):
        assert not self.is_local(), "A local ChildProcess does not have a ssh_connection"
        assert self.cp_started, "A ssh_connection will only be available after running execute_child_process()"
        return self.ssh_conn

    def get_queue_from_cp(self):
        assert self.set_up_port_for_structured_back_communication, "You did not specify a port to be set up. Don't expect structured info from child process"
        assert self.queue_from_cp, "The queue has not been set up yet. Did you start the child process before calling this function? Also make sure your implementation of child process has this functionality"
        return self.queue_from_cp

    def execute_child_process(self):
        '''
        Executes ChildProcess in a non-blocking manner. This returns immediately.
        This method returns a tuple (stdin, stdout, stderr) of the child process
        :return:
        '''
        if not self.is_local():
            self.ssh_conn = get_ssh_connection(self.ip_address)
        command = self.prepare_command(self.command)
        pid, stdin, stdout, stderr = self._run_command(command,get_pty=True)
        self._assign_pid(pid)
        self.cp_started = True
        return (stdin, stdout, stderr)

    def _run_command(self,command,get_pty=False):
        '''
        execute the given command
        :param command: string, to execute
        :return: (stdin, stdout, stderr)
        '''
        if self.local_process:
            if type(command) == list:
                sub = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            elif isinstance(command, string_types):
                shlexed_command = shlex.split(command)
                sub = subprocess.Popen(shlexed_command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            else:
                raise NotImplementedError()
            self.sub = sub
            stdin = sub.stdin
            stdout = sub.stdout
            stderr = sub.stderr
            pid = sub.pid
        else:
            stdin, stdout, stderr = self.ssh_conn.exec_command(command,get_pty=get_pty)
            pid = stdout.readline().strip()
        return (pid, stdin, stdout, stderr)

    def kill(self, signum=signal.SIGINT):
        '''
        Kills the process by
        (remote) sending 'kill -s signal pid' to the server.
        (local)  os.kill(pid, signal)
        default signal is SIGINT
        This call does not block. The success of killing the process
        needs to be determined by the user. E.g by calling is_alive().
        In case the process was killed, the ssh connection is terminated (if remote)
        :return:
        '''

        if not self.cp_started:
            print("Not started yet, no kill command will be sent")
            return
        if self.is_local():
            if check_pid(self.sub.pid):
                self.sub.send_signal(signum)
        else:
            if self.ssh_conn.get_transport() is None:
                return
            kill_command = "kill -s %s %s" %(signum, self.get_pid())
            self._run_command(kill_command)
            if self.ssh_conn.get_transport() is not None:
                if not self.is_alive() and not self.is_local():
                    self.ssh_conn.close()


    def is_alive(self):
        if not self.cp_started:
            return False
        if self.is_local():
            return self.sub.poll() == None
        else:
            if self.ssh_conn.get_transport() is None:
                return False
            else:
                command = "echo $$; exec ps -h -p %s"%self.get_pid()
                _,_,stdout,_ = self._run_command(command)
                return self.get_pid() in stdout.read()


class PythonChildProcess(ChildProcess):
    '''
    This ChildProcess is designed to spawn python processes.
    '''
    def __init__(self, ip_address, command, **kwargs):
        '''
        Creates a PythonChildProcess
        :param ip_address: The command will be executed at this ip_address
        :param command: the command to execute. Is assumed to be a python call
        :param name: optional name. If not set, will be process_i, with i a global counter
        :param take_care_of_deconstruct: If set to True, deconstruct() is registered at exit
        :param port_for_structured_back_communication: If set, the ip-address of this device as well as a specific port will be appended as arguments to the executed python command in the format --port=1234 --address=127.0.0.1
        The child process is responsible for reading these values out and communicating to it. If set, this child process will expose a queue on that address with get_queue_from_cp()
        :return:
        '''
        super(PythonChildProcess,self).__init__(ip_address=ip_address, command=command,**kwargs)

    def prepare_command(self,command):
        '''
        All the stuff that I need to prepare for the command to definitely work
        :param command:
        :return:
        '''
        if self.set_up_port_for_structured_back_communication:
            self.queue_from_cp, port = listen_on_port()
            if self.is_local():
                address = "127.0.0.1"
            else:
                address = get_local_ips()[-1]

        if self.is_local():
            home_dir = os.path.expanduser("~")
        else:
            _,_,stdout,_ = self._run_command("echo $$; exec echo ~")
            home_dir = stdout.read().strip()

        if type(command) == list:
            if self.set_up_port_for_structured_back_communication:
                command.append("--port=%i"%port)
                command.append("--address=%s"%address)
            if not self.local_process:
                command = [c.replace("python", self.get_extended_command(get_artemis_config_value(section=self.get_ip(), option="python", default_generator=lambda: sys.executable)), 1) if c.startswith("python") else c for c in command]
                command = [s.replace("~",home_dir) for s in command]
                command = " ".join([c for c in command])
            else:
                command = [c.strip("'") for c in command]
                command = [c.replace("python", sys.executable, 1) if c.startswith("python") else c for c in command]
                command = [s.replace("~",home_dir) for s in command]

        elif isinstance(command, string_types) and command.startswith("python"):
            if self.set_up_port_for_structured_back_communication:
                command += " --port=%i "%port
                command += "--address=%s"%address
            if not self.local_process:
                command = command.replace("python", self.get_extended_command(get_artemis_config_value(section=self.get_ip(), option="python",default_generator=lambda: sys.executable)), 1)
            else:
                command = command.replace("python", sys.executable)
            command = command.replace("~",home_dir)
        else:
            raise NotImplementedError()
        return command


def listen_on_port(port=7000):
    '''
    Sets up a thread that listens on a port forwards communication to a queue. Both, Queue and the port it listens on are returned
    :return: queue, port
    '''
    sock, port = get_socket("0.0.0.0", port=port)
    sock.listen(1)
    main_input_queue = queue.Queue()
    t = threading.Thread(target=handle_socket_accepts,args=(sock, main_input_queue, None,1))
    t.setDaemon(True)
    t.start()
    return main_input_queue, port


class RemotePythonProcess(ChildProcess):
    """
    Launch a python child process.
    """

    def __init__(self, function, ip_address, set_up_port_for_structured_back_communication=True, **kwargs):
        if ip_address=='localhost':
            ip_address="127.0.0.1"

        pickled_function = pickle_dumps_without_main_refs(function)
        encoded_pickled_function = base64.b64encode(pickled_function)

        remote_run_script_path = inspect.getfile(remote_function_run_script)
        if remote_run_script_path.endswith('pyc'):
            remote_run_script_path = remote_run_script_path[:-1]

        self.return_value_queue, return_port = listen_on_port(7000)
        all_local_ips = get_local_ips()
        return_address = "127.0.0.1" if ip_address in all_local_ips else all_local_ips[-1]
        command = [sys.executable, '-u', remote_run_script_path, encoded_pickled_function, return_address, str(return_port)]
        super(RemotePythonProcess, self).__init__(ip_address=ip_address, command=command, set_up_port_for_structured_back_communication=set_up_port_for_structured_back_communication,  **kwargs)

    def get_return_value(self, timeout=1):
        assert self.set_up_port_for_structured_back_communication, '{} has not been set up to send back a return value.'.format(self)
        serialized_out = self.return_value_queue.get(timeout=timeout)
        out = pickle.loads(serialized_out.dbplot_message)
        return out

class SlurmPythonProcess(RemotePythonProcess):
    def __init__(self, function, ip_address, set_up_port_for_structured_back_communication=True, slurm_kwargs={}, slurm_command="srun", **kwargs):
        '''

        :param function:
        :param ip_address:
        :param set_up_port_for_structured_back_communication:
        :param slurm_kwargs:
        :param kwargs:
        '''
        assert ip_address in get_local_ips(), "At the moment, we want you to start a slurm process only from localhost"
        assert slurm_command in ["srun"], "At the moment, we only support 'srun' for execution of slurm"
        super(SlurmPythonProcess,self).__init__(function, ip_address, set_up_port_for_structured_back_communication, **kwargs)
        self.slurm_kwargs = slurm_kwargs


    def prepare_command(self,command):
        '''
        All the stuff that I need to prepare for the command to definitely work
        :param command:
        :return:
        '''

        slurm_command = "srun"
        for k,v in self.slurm_kwargs.items():
            if k.startswith("--"):
                slurm_command += " %s=%s"%(k,v)
            elif k.startswith("-"):
                slurm_command += " %s %s"%(k,v)

        if isinstance(command, list):
            command = " ".join(pipes.quote(c) for c in command)

        final_command = " ".join((slurm_command,command))
        if self.is_local():
            return final_command
        else:
            raise NotImplementedError()

def pickle_dumps_without_main_refs(obj):
    """
    Yeah this is horrible, but it allows you to pickle an object in the main module so that it can be reloaded in another
    module.
    :param obj:
    :return:
    """
    currently_run_file = sys.argv[0]
    module_path = file_path_to_absolute_module(currently_run_file)
    try:
        pickle_str = pickle.dumps(obj, protocol=0)
    except:
        print("Using Dill")
        # TODO: @petered There is something very fishy going on here that I don't understand.
        import dill
        pickle_str = dill.dumps(obj, protocol=0)

    pickle_str = pickle_str.replace('__main__', module_path)  # Hack!
    return pickle_str


def pickle_dump_without_main_refs(obj, file_obj):
    string = pickle_dumps_without_main_refs(obj)
    file_obj.write(string)



import hashlib
import tempfile
from six.moves import StringIO
import gzip
import tarfile
from zipfile import ZipFile
from artemis.fileman.local_dir import get_artemis_data_path, make_dir, make_file_dir
from artemis.general.should_be_builtins import bad_value
import shutil
import os
from six.moves.urllib.request import urlopen

__author__ = 'peter'


def get_file(relative_name, url = None, data_transformation = None):

    relative_folder, file_name = os.path.split(relative_name)
    local_folder = get_artemis_data_path(relative_folder)

    try:  # Best way to see if folder exists already - avoids race condition between processes
        os.makedirs(local_folder)
    except OSError:
        pass

    full_filename = os.path.join(local_folder, file_name)

    if not os.path.exists(full_filename):
        assert url is not None, "No local copy of '%s' was found, and you didn't provide a URL to fetch it from" % (full_filename, )

        print('Downloading file from url: "%s"...' % (url, ))
        response = urlopen(url)
        data = response.read()
        print('...Done.')

        if data_transformation is not None:
            print('Processing downloaded data...')
            data = data_transformation(data)
        with open(full_filename, 'wb') as f:
            f.write(data)
    return full_filename


def get_file_in_archive(relative_path, subpath, url, force_extract = False):
    """
    Download a zip file, unpack it, and get the local address of a file within this zip (so that you can open it, etc).

    :param relative_path: Local name for the extracted folder.  (Zip file will be named this with the appropriate zip extension)
    :param url: Url of the zip file to download
    :param subpath: Path of the file relative to the zip folder.
    :param force_extract: Force the zip file to re-extract (rather than just reusing the extracted folder)
    :return: The full path to the file on your system.
    """
    local_folder_path = get_archive(relative_path=relative_path, url=url, force_extract=force_extract)
    local_file_path = os.path.join(local_folder_path, subpath)
    assert os.path.exists(local_file_path), 'Could not find the file "%s" within the extracted folder: "%s"' % (subpath, local_folder_path)
    return local_file_path


def get_archive(relative_path, url, force_extract=False, archive_type = None, force_download = False):
    """
    Download a compressed archive and extract it into a folder.

    :param relative_path: Local name for the extracted folder.  (Zip file will be named this with the appropriate zip extension)
    :param url: Url of the archive to download
    :param force_extract: Force the zip file to re-extract (rather than just reusing the extracted folder)
    :return: The full path to the extracted folder on your system.
    """

    local_folder_path = get_artemis_data_path(relative_path)

    assert archive_type in ('.tar.gz', '.zip', None)

    if force_download:
        shutil.rmtree(local_folder_path)

    if not os.path.exists(local_folder_path) or force_download:  # If the folder does not exist, download zip and extract.
        # (We also check force download here to avoid a race condition)
        response = urlopen(url)

        # Need to infer
        if archive_type is None:
            if url.endswith('.tar.gz'):
                archive_type = '.tar.gz'
            elif url.endswith('.zip'):
                archive_type = '.zip'
            else:
                # info = response.info()
                try:
                    # header = next(x for x in info.headers if x.startswith('Content-Disposition'))
                    header = response.headers['content-disposition']
                    original_file_name = next(x for x in header.split(';') if x.startswith('filename')).split('=')[-1].lstrip('"\'').rstrip('"\'')
                    archive_type = '.tar.gz' if original_file_name.endswith('.tar.gz') else '.zip' if original_file_name.endswith('.zip') else \
                        bad_value(original_file_name, 'Filename "%s" does not end with a familiar zip extension like .zip or .tar.gz' % (original_file_name, ))
                except StopIteration:
                    raise Exception("Could not infer archive type from user argument, url-name, or file-header.  Please specify archive type as either '.zip' or '.tar.gz'.")
        print('Downloading archive from url: "%s"...' % (url, ))
        data = response.read()
        print('...Done.')

        local_zip_path = local_folder_path + archive_type
        make_file_dir(local_zip_path)
        with open(local_zip_path, 'wb') as f:
            f.write(data)

        force_extract = True

    if force_extract:
        if archive_type == '.tar.gz':
            with tarfile.open(local_zip_path) as f:
                f.extractall(local_folder_path)
        elif archive_type == '.zip':
            with ZipFile(local_zip_path) as f:
                f.extractall(local_folder_path)
        else:
            raise Exception()

    return local_folder_path


def get_file_and_cache(url, data_transformation = None, enable_cache_write = True, enable_cache_read = True):

    _, ext = os.path.splitext(url)

    if enable_cache_read or enable_cache_write:
        hasher = hashlib.md5()
        hasher.update(url.encode('utf-8'))
        code = hasher.hexdigest()
        local_cache_path = os.path.join(get_artemis_data_path('caches'), code + ext)

    if enable_cache_read and os.path.exists(local_cache_path):
        return local_cache_path
    elif enable_cache_write:
        full_path = get_file(
            relative_name = os.path.join('caches', code+ext),
            url = url,
            data_transformation=data_transformation
            )
        return full_path
    else:
        return get_temp_file(url, data_transformation=data_transformation)


def get_temp_file(url, data_transformation = None):
    _, ext = os.path.splitext(url)
    tmp_file = tempfile.mktemp() + ext
    return get_file(tmp_file, url, data_transformation=data_transformation)


def unzip_gz(data):
    return gzip.GzipFile(fileobj = StringIO(data)).read()



import sys
import pytest
from artemis.experiments.decorators import experiment_function


def is_matplotlib_imported():
    return 'matplotlib' in sys.modules


@pytest.mark.skipif(True, reason='Test only works in isolation.  Matplotlib may be imported already when running with other tests in pytest.')
def test_matplotlib_dependency_is_soft():

    assert not is_matplotlib_imported()

    @experiment_function
    def my_test_exp(a=1, b=3):
        print('aaa')
        return a*b

    assert not is_matplotlib_imported()
    my_test_exp.run(save_figs = False)  # When Running an experiment

    my_test_exp.browse(command='q')  # Also in UI

    assert not is_matplotlib_imported()
    from matplotlib import pyplot as plt
    assert is_matplotlib_imported()


if __name__ == '__main__':
    test_matplotlib_dependency_is_soft()



from functools import partial

import sys
from pytest import raises
from artemis.general.functional import infer_arg_values, get_partial_chain, infer_derived_arg_values


def test_get_full_args():

    def func(a, b):
        return a+b

    assert list(infer_arg_values(func, a=1, b=3).items()) == [('a', 1), ('b', 3)]
    assert list(infer_arg_values(func, 1, b=3).items()) == [('a', 1), ('b', 3)]
    assert list(infer_arg_values(func, 1, 3).items()) == [('a', 1), ('b', 3)]
    with raises(AssertionError):  # AssertionError: Arguments ('a', 'b') require values but are not given any.
        infer_arg_values(func, )
    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_arg_values(func, b=3)
    with raises(AssertionError):  # AssertionError: The set of argument names to the function: ('a', 'b') must match the set of arguments given: ('a', 'b', 'c')
        infer_arg_values(func, a=1, b=3, c=5)
    with raises(AssertionError):  # AssertionError: Arguments ('b',) require values but are not given any.
        infer_arg_values(func, 1, c=5)
    with raises(AssertionError):  # AssertionError: You provided 3 arguments, but the function only takes 2
        print(infer_arg_values(func, 1, 2, 5))
    with raises(AssertionError):  # AssertionError: Arguments ('b',) have been defined multiple times: (('a', 1), ('b', 2), ('b', 5))
        infer_arg_values(func, 1, 2, b=5)

    def func_with_defaults(a, b=4):
        return a+b
    assert list(infer_arg_values(func_with_defaults, a=1).items()) == [('a', 1), ('b', 4)]
    assert list(infer_arg_values(func_with_defaults, a=1, b=3).items()) == [('a', 1), ('b', 3)]
    assert list(infer_arg_values(func_with_defaults, 1).items()) == [('a', 1), ('b', 4)]
    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_arg_values(func_with_defaults, )
    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_arg_values(func_with_defaults, b=3)
    with raises(AssertionError):  # AssertionError: The set of argument names to the function: ('a', 'b') must match the set of arguments given: ('a', 'b', 'c')
        infer_arg_values(func_with_defaults, a=1, b=3, c=5)
    with raises(AssertionError):  # AssertionError: The set of argument names to the function: ('a', 'b') must match the set of arguments given: ('a', 'b', 'c')
        infer_arg_values(func_with_defaults, 1, c=5)
    with raises(AssertionError):  # AssertionError: You provided 3 arguments, but the function only takes 2
        print(infer_arg_values(func_with_defaults, 1, 2, 5))
    with raises(AssertionError):  # AssertionError: Arguments ('b',) have been defined multiple times: (('a', 1), ('b', 2), ('b', 5))
        infer_arg_values(func_with_defaults, 1, 2, b=5)

    def func_with_kwargs(a, b=4, **kwargs):
        return a+b
    assert list(infer_arg_values(func_with_kwargs, a=1).items()) == [('a', 1), ('b', 4)]
    assert list(infer_arg_values(func_with_kwargs, a=1, b=3).items()) == [('a', 1), ('b', 3)]
    assert list(infer_arg_values(func_with_kwargs, 1).items()) == [('a', 1), ('b', 4)]
    list(infer_arg_values(func_with_kwargs, a=1, b=3, c=5)) == [('a', 1), ('b', 3), ('c', 5)]
    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_arg_values(func_with_kwargs, )
    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_arg_values(func_with_kwargs, b=3)
    with raises(AssertionError):  # AssertionError: You provided 3 arguments, but the function only takes 2
        print(list(infer_arg_values(func_with_kwargs, 1, 2, 5)))
    with raises(AssertionError):  # AssertionError: Arguments ('b',) have been defined multiple times: (('a', 1), ('b', 2), ('b', 5))
        infer_arg_values(func_with_kwargs, 1, 2, b=5)


def test_get_partial_chain():
    def f(a, b):
        return a+b
    g = partial(f, b=3)
    h = partial(g, a=2)

    if sys.version_info < (3, 0):  # Python 2.X
        assert [f, g, h] == get_partial_chain(h)
    else: # Python 3.X
        base, part = get_partial_chain(h)
        assert base is f
        assert part.keywords == {'a': 2, 'b': 3}


def test_get_derived_function_args():
    def f(a, b=1):
        return a+b
    g = partial(f, a=2)
    h = partial(g, b=3)
    j = partial(h, c=4)

    with raises(AssertionError):  # AssertionError: Arguments ('a',) require values but are not given any.
        infer_derived_arg_values(f)
    assert list(infer_derived_arg_values(g).items()) == [('a', 2), ('b', 1)]
    assert list(infer_derived_arg_values(h).items()) == [('a', 2), ('b', 3)]
    with raises(AssertionError):
        infer_derived_arg_values(j)

    assert f(**infer_derived_arg_values(g)) == g()
    assert f(**infer_derived_arg_values(h)) == h()


if __name__ == '__main__':
    test_get_full_args()
    test_get_partial_chain()
    test_get_derived_function_args()



'''
This code is taken from https://github.com/paramiko/paramiko/blob/master/demos/forward.py and adapted
'''

from __future__ import print_function
import select

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer


class ForwardServer(SocketServer.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(SocketServer.BaseRequestHandler):
    def handle(self):
        try:
            chan = self.ssh_transport.open_channel('direct-tcpip',
                                                   (self.chain_host, self.chain_port),
                                                   self.request.getpeername())
        except Exception as e:
            print('Incoming request to %s:%d failed: %s' % (self.chain_host,
                                                            self.chain_port,
                                                            repr(e)))
            return
        if chan is None:
            print('Incoming request to %s:%d was rejected by the SSH server.' %
                  (self.chain_host, self.chain_port))
            return

        print('Connected!  Tunnel open %r -> %r -> %r' % (self.request.getpeername(),
                                                          chan.getpeername(), (self.chain_host, self.chain_port)))
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        print('Tunnel closed from %r' % (peername,))


def forward_tunnel(local_port, remote_host, remote_port, ssh_conn):
    # this is a little convoluted, but lets me configure things for the Handler
    # object.  (SocketServer doesn't give Handlers any way to access the outer
    # server normally.)
    class SubHandler(Handler):
        chain_host = remote_host
        chain_port = remote_port
        ssh_transport = ssh_conn.get_transport()

    try:
        ForwardServer(('', local_port), SubHandler).serve_forever()
    except SocketServer.socket.error as exc:
        if exc.args[0] == 48:
            # port forwarding on this device already set up?
            pass
        else:
            raise



#!/usr/bin/python
import base64

from artemis.remote.utils import one_time_send_to
import sys
import pickle
"""
This file is called from within a ChildProcess
"""

if __name__ == '__main__':
    _, encoded_pickled_function, return_address, return_port = sys.argv
    pickled_function = base64.b64decode(encoded_pickled_function)
    try:
        func = pickle.loads(pickled_function)
    except:
        print("Using dill to unpickle")
        import dill
        func = dill.loads(pickled_function)
    result = func()
    pickled_result = pickle.dumps(result, protocol = pickle.HIGHEST_PROTOCOL)
    try:
        one_time_send_to(return_address, return_port, pickled_result)
    except:
        pass



from scipy.misc.pilutil import imresize
import numpy as np
__author__ = 'peter'


def resize_while_preserving_aspect_ratio(im, x_dim=None, y_dim=None):
    """
    Resize an image, while preserving the aspect ratio.  For this you need to specify either x_dim or y_dim.

    :param im: The image: a 2D or 3D array.
    :param x_dim: An integer indicating the desired size, or None, to leave it loose.
    :param y_dim: An integer indicating the desired size, or None, to leave it loose.
    :return: A new image whose x_dim or y_dim matches the constraint
    """
    assert not (x_dim is None and y_dim is None), 'You can not leave both constraints at None!'

    x_dim = float('inf') if x_dim is None else x_dim
    y_dim = float('inf') if y_dim is None else y_dim

    box_aspect_ratio = x_dim/float(y_dim)
    image_aspect_ratio = im.shape[1] / float(im.shape[0])
    if image_aspect_ratio > box_aspect_ratio:  # Active constraint is width
        return imresize(im, size=(int(x_dim/image_aspect_ratio+.5), x_dim))
    else:  # Active constraint is height
        return imresize(im, size=(y_dim, int(y_dim*image_aspect_ratio+.5)))


def equalize_image_dims(list_of_images, x_dim = None, y_dim = None):
    """
    Resize images so that they match roughly in size although their aspect ratio will be preserved.
    :param list_of_images: A list of numpy arrays representing images (2D or 3D arrays)
    :param size: A 2-tuple specifying the desired (y_size, x_size).
        Each of (y_size, x_size) can be:
        - An integar, meaning that this axis of the image will remain equal or smaller than this number of pixels.
        - None, meaning that there is no constraint along this axis (e.g. (224, None) just states that the image will be
          scaled to 224 pixels in the vertical direction - the horizontal will be whatever size is needed to maintain
          the aspect ratio.
        - 'max': Meaning that we take the largest image size along this axis.
        - 'min': Meaning what we take the largest image size along this axis.

        The image will then be scaled so that the image size remains inside this box (although, unless the aspect ratio
        matches exactly, one dimension will be smaller).

    :return: Another list of images.
    """
    assert not (x_dim is None and y_dim is None), 'You can not leave both constraints at None!'
    if len(list_of_images)==0:
        return []
    x_dim = max(im.shape[1] for im in list_of_images) if x_dim=='max' else \
        min(im.shape[1] for im in list_of_images) if x_dim=='min' else \
        x_dim
    y_dim = max(im.shape[0] for im in list_of_images) if y_dim=='max' else \
        min(im.shape[0] for im in list_of_images) if y_dim=='min' else \
        y_dim
    new_list_of_images = [resize_while_preserving_aspect_ratio(im, x_dim=x_dim, y_dim=y_dim) for im in list_of_images]
    return new_list_of_images


def resize_and_crop(im, width, height):
    im_aspect = float(im.shape[0])/im.shape[1]
    new_aspect = float(height)/width
    if im_aspect > new_aspect:  # Need to chop the top and bottom
        new_height = int(width*im_aspect)
        resized_im = imresize(im, (new_height, width))
        start = (new_height-height)/2
        output_im = resized_im[start:start+height, :]
    else:  # Need to chop the left and right.
        new_width = int(height/im_aspect)
        resized_im = imresize(im, (height, new_width))
        start = (new_width-width)/2
        output_im = resized_im[:, start:start+width]
    assert output_im.shape[:2] == (height, width)
    return output_im


def resize_image(im, width=None, height=None, mode='squeeze'):
    assert isinstance(im, np.ndarray) and im.ndim in (2, 3)
    if mode == 'squeeze':
        im = imresize(im, size=(height, width))
    elif mode == 'preserve_aspect':
        im = resize_while_preserving_aspect_ratio(im, x_dim=width, y_dim=height)
    elif mode == 'crop':
        current_height, current_width = im.shape[:2]
        assert height>=height and width>=width, "Crop size must be smaller than image size"
        row_start = (current_height-height)/2
        col_start = (current_width-width)/2
        im = im[..., row_start:row_start+224, col_start:col_start+224, :]
    elif mode in ('resize_and_crop', 'scale_crop'):
        assert height is not None and width is not None, "You need to specify both height and width. for 'scale_crop' mode"
        return resize_and_crop(im, width=width, height=height)
    else:
        raise Exception("Unknown resize mode: '{}'".format(mode))
    return im


def get_dark_edge_slice(im, cut_edges_thresh=0):
    vnonzero = np.flatnonzero(im.mean(axis=2).mean(axis=1)>cut_edges_thresh)
    hnonzero = np.flatnonzero(im.mean(axis=2).mean(axis=0)>cut_edges_thresh)
    edge_crops = slice(vnonzero[0], vnonzero[-1]+1), slice(hnonzero[0], hnonzero[-1]+1)
    return edge_crops


def cut_dark_edges(im, slices = None, cut_edges_thresh=0):
    if slices is None:
        slices = get_dark_edge_slice(im, cut_edges_thresh=cut_edges_thresh)
    y_slice, x_slice = slices
    return im[y_slice, x_slice]



import argparse
import logging
import os
import pickle
import shlex
import shutil
from collections import OrderedDict
from functools import partial
from multiprocessing import Process

from six.moves import input
from tabulate import tabulate

from artemis.experiments.experiment_management import deprefix_experiment_ids, \
    RecordSelectionError, run_multiple_experiments_with_slurm
from artemis.experiments.experiment_management import get_experient_to_record_dict
from artemis.experiments.experiment_management import (pull_experiments, select_experiments, select_experiment_records,
                                                       select_experiment_records_from_list, interpret_numbers,
                                                       run_multiple_experiments)
from artemis.experiments.experiment_record import ExpStatusOptions
from artemis.experiments.experiment_record import (get_all_record_ids, clear_experiment_records,
                                                   load_experiment_record, ExpInfoFields)
from artemis.experiments.experiment_record_view import (get_record_full_string, get_record_invalid_arg_string,
                                                        print_experiment_record_argtable, get_oneline_result_string,
                                                        compare_experiment_records)
from artemis.experiments.experiment_record_view import show_record, show_multiple_records
from artemis.experiments.experiments import load_experiment, get_nonroot_global_experiment_library
from artemis.fileman.local_dir import get_artemis_data_path
from artemis.general.display import IndentPrint, side_by_side, truncate_string, surround_with_header, format_duration, format_time_stamp
from artemis.general.hashing import compute_fixed_hash
from artemis.general.mymath import levenshtein_distance
from artemis.general.should_be_builtins import all_equal, insert_at, izip_equal, separate_common_items, bad_value

try:
    import readline  # Makes input() behave like interactive shell.
    # http://stackoverflow.com/questions/15416054/command-line-in-python-with-history
except:
    pass  # readline not available


try:
    from enum import Enum
except ImportError:
    raise ImportError("Failed to import the enum package. This was added in python 3.4 but backported back to 2.4.  To install, run 'pip install --upgrade pip enum34'")


def _warn_with_prompt(message= None, prompt = 'Press Enter to continue or q then Enter to quit', use_prompt=True):
    if message:
        print(message)
    if use_prompt:
        resp = input('({}) >> '.format(prompt))
        out = resp.strip().lower()
        if out=='q':
            quit()
        else:
            return out


def browse_experiments(command=None, **kwargs):
    """
    Browse Experiments

    :param command: Optionally, a string command to pass directly to the UI.  (e.g. "run 1")
    :param root_experiment: The Experiment whose (self and) children to browse
    :param catch_errors: Catch errors that arise while running experiments
    :param close_after: Close after issuing one command.
    :param just_last_record: Only show the most recent record for each experiment.
    :param view_mode: How to view experiments {'full', 'results'} ('results' leads to a narrower display).
    :param raise_display_errors: Raise errors that arise when displaying the table (otherwise just indicate that display failed in table)
    :param run_args: A dict of named arguments to pass on to Experiment.run
    :param keep_record: Keep a record of the experiment after running.
    :param truncate_result_to: An integer, indicating the maximum length of the result string to display.
    :param cache_result_string: Cache the result string (useful when it takes a very long time to display the results
        when opening up the menu - often when results are long lists).
    """
    browser = ExperimentBrowser(**kwargs)
    browser.launch(command=command)


class ExperimentBrowser(object):

    QUIT = 'Quit'
    REFRESH = 'Refresh'
    HELP_TEXT = """
This program lists the experiments that you have defined (referenced by E#) alongside the records of console output,
plots, results, referenced by (E#.R# - for example 4.1) created by running these experiments.  Command examples:

> 4                   Run experiment 4
> run 4               Run experiment 4
> run 4-6             Run experiment 4, 5, and 6
> run all             Run all experiments
> run unfinished      Run all experiments which have not yet run to completion.
> run 4-6 -s          Run experiments 4, 5, and 6 in sequence, and catch all errors.
> run 4-6 -e          Run experiments 4, 5, and 6 in sequence, and stop on errors
> run 4-6 -p          Run experiments 4, 5, and 6 in parallel processes, and catch all errors.
> run 4-6 -p2         Run experiments 4, 5, and 6 in parallel processes, using up to 2 processes at a time.
> call 4              Call experiment 4 (like running, but doesn't save a record)
> filter 4-6          Just show experiments 4-6 and their records
> filter has:xyz      Just show experiments with "xyz" in the name and their records
> filter 1diff:3      Just show all experiments that have no more than 1 argument different from experiment 3.
> filter -            Clear the filter and show the full list of experiments
> filterrec last      Just show the last record of each experiment
> filterrec finished  Just show completed records
> filterrec ~finished Just show non-completed experiments
> filterrec finished>last   Just show the last finished runs of each experiment
> results 4-6         View the results experiments 4, 5, 6
> view results        View just the columns for experiment name and result
> view full           View all columns (the default view)
> show 4              Show the output from the last run of experiment 4 (if it has been run already).
> show 4-6            Show the output of experiments 4,5,6 together.
> records             Browse through all experiment records.
> compare 4.1,5.3     Print a table comparing the arguments and results of records 4.1 and 5.3.
> selectexp 4-6       Show the list of experiments (and their records) selected by the "experiment selector" "4-6" (see below for possible experiment selectors)
> selectrec 4-6       Show the list of records selected by the "record selector" "4-6" (see below for possible record selectors)
> sidebyside 4.1,5.3  Display the output of record from experiments 4.1,5.3 side by side.
> delete 4.1          Delete record 1 of experiment 4
> delete unfinished   Delete all experiment records that have not run to completion
> delete 4-6          Delete all records from experiments 4, 5, 6.  You will be asked to confirm the deletion.
> pull 1-4 machine1   Pulls records from experiments 1,2,3,4 from machine1 (requires some setup, see artemis.remote.remote_machines.py)
> q                   Quit.
> r                   Refresh list of experiments.
> clearcache          Clear the cached display of experiment records in the UI (caching is used only if cache_result_string==True)

Commands 'run', 'call', 'filter', 'pull', '1diff', 'selectexp' allow you to select experiments.  You can select
experiments in the following ways:

    Experiment
    Selector        Action
    --------------- ---------------------------------------------------------------------------------------------------
    4               Select experiment #4
    4-6             Select experiments 4, 5, 6
    all             Select all experiments
    unfinished      Select all experiment for which there are no records of it being run to completion.
    invalid         Select all experiments where all records were made before arguments to the experiment have changed
    has:xyz         Select all experiments with the string "xyz" in their names
    hasnot:xyz      Select all experiments without substring "xyz" in their names
    1diff:3         Select all experiments who have no more than 1 argument which is different from experiment 3's arguments.

Commands 'results', 'show', 'records', 'compare', 'sidebyside', 'selectrec', 'filterrec', 'delete' allow you to specify a range of
experiment records.  You can specify records in the following ways:

    Record
    Selector        Action
    --------------- ---------------------------------------------------------------------------------------------------
    4.2             Select record 2 for experiment 4
    4               Select all records for experiment 4
    4-6             Select all records for experiments 4, 5, 6
    4.2-5           Select records 2, 3, 4, 5 for experiment 4
    4.3,4.4         Select records 4.3, 4.4
    all             Select all records
    old             Select all records that are not the the most recent run for that experiment
    finished        Select all records that have not run to completion
    invalid         Select all records for which the arguments to their experiments have changed since they were run
    errors          Select all records that ended in error
    ~invalid        Select all records that are not invalid (the '~' can be used to negate any of the above)
    invalid|errors  Select all records that are invalid or ended in error (the '|' can be used to "or" any of the above)
    invalid&errors  Select all records that are invalid and ended in error (the '&' can be used to "and" any of the above)
    finished@last   Select the last finished record of each experiment (the '@' can be used to cascade any of the above)
"""

    def __init__(self, root_experiment = None, catch_errors = True, close_after = False, filterexp=None, filterrec = None,
            view_mode ='full', raise_display_errors=False, run_args=None, keep_record=True, truncate_result_to=100,
            ignore_valid_keys=(), cache_result_string = False, slurm_kwargs={}, remove_prefix = None, display_format='nested',
            show_args=False, catch_selection_errors=True, max_width=None, table_package = 'tabulate'):
        """
        :param root_experiment: The Experiment whose (self and) children to browse
        :param catch_errors: Catch errors that arise while running experiments
        :param close_after: Close after issuing one command.
        :param filterexp: Filter the experiments with this selection (see help for how to use)
        :param filterrec: Filter the experiment records with this selection (see help for how to use)
        :param view_mode: How to view experiments {'full', 'results'} ('results' leads to a narrower display).
        :param raise_display_errors: Raise errors that arise when displaying the table (otherwise just indicate that display failed in table)
        :param run_args: A dict of named arguments to pass on to Experiment.run
        :param keep_record: Keep a record of the experiment after running.
        :param truncate_result_to: An integer, indicating the maximum length of the result string to display.
        :param ignore_valid_keys: When checking whether arguments are valid, ignore arguments with names in this list
        :param cache_result_string: Cache the result string (useful when it takes a very long time to display the results
            when opening up the menu - often when results are long lists).
        :param slurm_kwargs:
        :param remove_prefix: Remove the common prefix on the experiment ids in the display.
        :param display_format: How experements and their records are displayed: 'nested' or 'flat'.  'nested' might be
            better for narrow console outputs.
        """

        if run_args is None:
            run_args = {}
        if 'keep_record' not in run_args:
            run_args['keep_record'] = keep_record
        if remove_prefix is None:
            remove_prefix = display_format=='flat'
        self.root_experiment = root_experiment
        self.close_after = close_after
        self.catch_errors = catch_errors
        self.exp_record_dict = None
        self.raise_display_errors = raise_display_errors
        self.view_mode = view_mode
        self._filter = filterexp
        self._filterrec = filterrec
        self.run_args = {} if run_args is None else run_args
        self.truncate_result_to = truncate_result_to
        self.cache_result_string = cache_result_string
        self.ignore_valid_keys = ignore_valid_keys
        self.slurm_kwargs = slurm_kwargs
        self.remove_prefix = remove_prefix
        self.display_format = display_format
        self.show_args = show_args
        self.catch_selection_errors = catch_selection_errors
        self.max_width = max_width
        self.table_package = table_package

    def _reload_record_dict(self):
        names = get_nonroot_global_experiment_library().keys()
        if self.root_experiment is not None:
            # We could just go [ex.name for ex in self.root_experiment.get_all_variants(include_self=True)]
            # but we want to preserve the order in which experiments were created
            descendents_of_root = set(ex.name for ex in self.root_experiment.get_all_variants(include_self=True))
            names = [name for name in names if name in descendents_of_root]
        all_experiments = get_experient_to_record_dict(names)
        return all_experiments

    def _filter_record_dict(self, all_experiments):
        # Apply filters and display Table:
        if self._filter is not None:
            try:
                all_experiments = OrderedDict((exp_name, all_experiments[exp_name]) for exp_name in select_experiments(self._filter, all_experiments))
            except RecordSelectionError as err:
                old_filter = self._filter
                self._filter = None
                raise RecordSelectionError("Failed to apply experiment filter: '{}' because {}.  Removing filter.".format(old_filter, err))
        if self._filterrec is not None:
            try:
                all_experiments = select_experiment_records(self._filterrec, all_experiments, load_records=False, flat=False)
            except RecordSelectionError as err:
                old_filterrec = self._filterrec
                self._filterrec = None
                raise RecordSelectionError("Failed to apply record filter: '{}' because {}.  Removing filter.".format(old_filterrec, err))
        return all_experiments

    def launch(self, command=None):

        func_dict = {
            'run': self.run,
            'test': self.test,
            'show': self.show,
            'call': self.call,
            'kill': self.kill,
            'selectexp': self.selectexp,
            'selectrec': self.selectrec,
            'view': self.view,
            'h': self.help,
            'filter': self.filter,
            'filterrec': self.filterrec,
            'displayformat': self.displayformat,
            'explist': self.explist,
            'sidebyside': self.side_by_side,
            'argtable': self.argtable,
            'compare': self.compare,
            'delete': self.delete,
            'errortrace': self.errortrace,
            'q': self.quit,
            'records': self.records,
            'pull': self.pull,
            'clearcache': clear_ui_cache,
            }

        display_again = True
        while True:

            if display_again:
                    all_experiments = self._reload_record_dict()
                    try:
                        self.exp_record_dict = self._filter_record_dict(all_experiments)
                    except RecordSelectionError as err:
                        _warn_with_prompt(str(err), use_prompt=self.catch_selection_errors)
                        if not self.catch_selection_errors:
                            raise
                        else:
                            continue
                    print(self.get_experiment_list_str(self.exp_record_dict))
                    if self._filter is not None or self._filterrec is not None:
                        print('[Showing {}/{} experiments and {}/{} records after Experiment Filter: "{}" and Record Filter: "{}"]'.format(
                            len(self.exp_record_dict), len(all_experiments),
                            sum(len(v) for v in self.exp_record_dict.values()), sum(len(v) for v in all_experiments.values()),
                            self._filter, self._filterrec
                            ))

            # Get command from user
            if command is None:
                user_input = input('Enter command or experiment # to run (h for help) >> ').strip()
            else:
                user_input = command
                command = None

            display_again = True
            out = None
            # Respond to user input
            with IndentPrint():
                try:
                    split = user_input.split(' ')
                    if len(split)==0:
                        continue
                    cmd = split[0]
                    args = split[1:]

                    if cmd in ('', 'r'): # Refresh
                        continue
                    elif interpret_numbers(cmd) is not None:  # If you've listed a number, you implicitely call run
                        args = [cmd]+args
                        cmd = 'run'

                    if cmd in func_dict:
                        out = func_dict[cmd](*args)
                        display_again = False
                    else:
                        edit_distances = [levenshtein_distance(cmd, other_cmd) for other_cmd in func_dict.keys()]
                        min_distance = min(edit_distances)
                        closest = func_dict.keys()[edit_distances.index(min_distance)]
                        suggestion = ' Did you mean "{}"?.  '.format(closest) if min_distance<=2 else ''
                        if self.close_after:
                            raise Exception('Unrecognised command: "{}"'.format(cmd))
                        else:
                            print('Unrecognised command: "{}". {}'.format(cmd, suggestion))
                        display_again = False
                    if out is self.QUIT or self.close_after:
                        break
                    elif out is self.REFRESH:
                        display_again = True

                except Exception as err:
                    print ("CAUGHT: {}".format(err))
                    if self.catch_errors or (isinstance(err, RecordSelectionError) and self.catch_selection_errors):
                        res = input('Enter "e" to view the stacktrace, or anything else to continue.')
                        if res.strip().lower() == 'e':
                            raise
                        display_again = True
                    else:
                        raise

    def get_experiment_list_str(self, exp_record_dict):
        headers = {
            'full': [ExpRecordDisplayFields.RUNS, ExpRecordDisplayFields.DURATION, ExpRecordDisplayFields.STATUS, ExpRecordDisplayFields.ARGS_CHANGED, ExpRecordDisplayFields.RESULT_STR, ExpRecordDisplayFields.NOTES],
            'results': [ExpRecordDisplayFields.RESULT_STR]
            }[self.view_mode]

        if self.remove_prefix:
            deprefixed_ids = deprefix_experiment_ids(exp_record_dict.keys())
            exp_record_dict = OrderedDict((k, v) for k, v in zip(deprefixed_ids, exp_record_dict.values()))

        row_func = _get_record_rows_cached if self.cache_result_string else _get_record_rows
        header_names = [h.value for h in headers]

        def remove_notes_if_no_notes(_record_rows, _record_headers):
            notes_column_index = _record_headers.index(ExpRecordDisplayFields.NOTES.value) if ExpRecordDisplayFields.NOTES.value in _record_headers else None
            # Remove the notes column if there are no notes!
            if notes_column_index is not None and all(row[notes_column_index]=='' or row[notes_column_index]=='-' for row in _record_rows):
                new_rows = []
                for row in _record_rows:
                    new_rows.append(row[:notes_column_index]+row[notes_column_index+1:])
                new_headers = _record_headers[:notes_column_index]+_record_headers[:notes_column_index+1:]
            else:
                new_rows = _record_rows
                new_headers = _record_headers
            return new_rows, new_headers

        if self.display_format=='nested':  # New Display mode

            if self.show_args:
                _, argdiff = separate_common_items([load_experiment(ex).get_args().items() for ex in exp_record_dict])
                argdiff = {k: args for k, args in izip_equal(exp_record_dict.keys(), argdiff)}
            # Build a list of experiments and a list of records.
            full_headers = ['#']+header_names
            record_rows = []
            experiment_rows = []
            experiment_row_ixs = []
            counter = 1  # Start at 2 because record table has the headers.
            for i, (exp_id, record_ids) in enumerate(exp_record_dict.items()):
                experiment_row_ixs.append(counter)
                exp_identifier = exp_id if not self.show_args else ','.join('{}={}'.format(k, v) for k, v in argdiff[exp_id])
                experiment_rows.append([i, exp_identifier])
                for j, record_id in enumerate(record_ids):
                    record_rows.append([j]+row_func(record_id, headers, raise_display_errors=self.raise_display_errors, truncate_to=self.truncate_result_to, ignore_valid_keys=self.ignore_valid_keys))
                    counter+=1
            record_rows, full_headers = remove_notes_if_no_notes(record_rows, full_headers)
            # Merge the experiments table and record table.

            if self.table_package=='tabulate':
                record_table_rows = tabulate(record_rows, headers=full_headers, tablefmt="pipe").split('\n')
                del record_table_rows[1]  # Get rid of that silly line.
                experiment_table_rows = tabulate(experiment_rows, numalign='left').split('\n')[1:-1]  # First and last are just borders
                longest_row = max(max(len(r) for r in record_table_rows), max(len(r) for r in experiment_table_rows)+4) if len(record_table_rows)>0 else 0
                record_table_rows = [r if len(r)==longest_row else r[:-1] + ' '*(longest_row-len(r)) + r[-1] for r in record_table_rows]
                experiment_table_rows = [('=' if i==0 else '-')*longest_row+'\n'+r + ' '*(longest_row-len(r)-1)+'|' for i, r in enumerate(experiment_table_rows)]
                all_rows = [surround_with_header('Experiments', width=longest_row, char='=')] + insert_at(record_table_rows, experiment_table_rows, indices=experiment_row_ixs) + ['='*longest_row]
                table = '\n'.join(all_rows)
            else:
                raise NotImplementedError(self.table_package)

        elif self.display_format=='flat':  # Display First record on same row
            full_headers = ['E#', 'R#', 'Experiment']+header_names
            rows = []
            for i, (exp_id, record_ids) in enumerate(exp_record_dict.items()):
                if len(record_ids)==0:
                    rows.append([str(i), '', exp_id, '<No Records>'] + ['-']*(len(headers)-1))
                else:
                    for j, record_id in enumerate(record_ids):
                        rows.append([str(i) if j==0 else '', j, exp_id if j==0 else '']+row_func(record_id, headers, raise_display_errors=self.raise_display_errors, truncate_to=self.truncate_result_to, ignore_valid_keys=self.ignore_valid_keys))
            assert len(rows[0])==len(full_headers)
            rows, full_headers = remove_notes_if_no_notes(rows, full_headers)

            if self.table_package == 'pretty_table':
                from prettytable.prettytable import PrettyTable
                table = str(PrettyTable(rows, field_names=full_headers, align='l', max_table_width=self.max_width))
            elif self.table_package == 'tabulate':
                table = tabulate(rows, headers=full_headers)
            else:
                raise NotImplementedError(self.table_package)
        else:
            raise NotImplementedError(self.display_format)

        return table

    def run(self, *args):

        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', help='A selection of experiments to run.  Examples: "3" or "3-5", or "3,4,5"')
        parser.add_argument('-p', '--parallel', default=False, nargs='*')
        parser.add_argument('-n', '--note')
        parser.add_argument('-e', '--raise_errors', default='single', nargs='*', help='By default, error are only raised if a single experiment is run.  Set "-e" to always rays errors.  "-e 0" to never raise errors.')
        parser.add_argument('-d', '--display_results', default=False, action = "store_true")
        parser.add_argument('-s', '--slurm', default=False, action = "store_true", help='Run with slurm')
        args = parser.parse_args(args)

        n_processes = \
            None if args.parallel is False else \
            'all' if len(args.parallel)==0 else \
            int(args.parallel[0]) if len(args.parallel)==1 else \
            bad_value(args.parallel, '-p can have 0 or 1 arguments.  Got: {}'.format(args.parallel))

        ids = select_experiments(args.user_range, self.exp_record_dict)

        # Raise errors if:
        # -e
        # -e 1
        # No arg, and only 1 experiment running
        raise_errors = (len(args.raise_errors)==0 or (len(args.raise_errors)==1 and args.raise_errors[0]=='1') or (args.raise_errors == 'single' and len(ids)==1))

        if args.slurm:
            run_multiple_experiments_with_slurm(
                experiments=[load_experiment(eid) for eid in ids],
                n_parallel = n_processes,
                raise_exceptions = raise_errors,
                run_args=self.run_args,
                slurm_kwargs=self.slurm_kwargs
                )
        else:
            exp_names = list(self.exp_record_dict.keys())
            run_multiple_experiments(
                experiments=[load_experiment(eid) for eid in ids],
                prefixes=[exp_names.index(eid) for eid in ids],
                parallel=n_processes,
                raise_exceptions = raise_errors,
                run_args=self.run_args,
                notes=(args.note, ) if args.note is not None else (),
                display_results=args.display_results
                )

        result = _warn_with_prompt('Finished running {} experiment{}.'.format(len(ids), '' if len(ids)==1 else 's'),
                use_prompt=not self.close_after,
                prompt='Press Enter to Continue, or "q" then Enter to Quit')
        if result=='q':
            quit()

    def test(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier).test()

    def help(self):
        _warn_with_prompt(self.HELP_TEXT, prompt = 'Press Enter to exit help.', use_prompt=not self.close_after)

    def show(self, *args):
        """
        :param user_range:  A range specifying the record
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', help='A selection of experiment records to show. ')
        parser.add_argument('-l', '--last', action='store_true', help='Just select the last record from the list of experiments.')
        parser.add_argument('-r', '--results', default=False, action = "store_true", help="Only show records with results.")
        parser.add_argument('-o', '--original', default=False, action = "store_true", help="Use the original default show function: show_record")
        parser.add_argument('-p', '--process', default=False, action = "store_true", help="Launch show in new process")
        args = parser.parse_args(args)

        user_range = args.user_range
        if args.results:
            user_range += '@result'
        if args.last:
            user_range += '@last'

        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)

        func = show_record if args.original else None
        if args.process:
            Process(target=partial(show_multiple_records, records, func)).start()
        else:
            show_multiple_records(records, func)
        _warn_with_prompt(use_prompt=False)

    def compare(self, *args):
        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', help='A selection of experiment records to compare.  Examples: "3" or "3-5", or "3,4,5"')
        parser.add_argument('-l', '--last', default=False, action = "store_true", help="Use this flag if you want to select Experiments instead of Experiment Records, and just show the last completed.")
        parser.add_argument('-r', '--results', default=False, action = "store_true", help="Only compare records with results.")
        parser.add_argument('-o', '--original', default=False, action = "store_true", help="Use original compare funcion")

        args = parser.parse_args(args)

        user_range = args.user_range if not args.last else args.user_range + '@result@last'
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        if args.results:
            records = [rec for rec in records if rec.has_result()]
        if len(records)==0:
            raise RecordSelectionError('No records were selected with "{}"'.format(args.user_range))

        if args.original:
            func = compare_experiment_records
        else:
            compare_funcs = [rec.get_experiment().compare for rec in records]
            assert all_equal(compare_funcs), "Your records have different comparison functions - {} - so you can't compare them".format(set(compare_funcs))
            func = compare_funcs[0]

        # The following could be used to launch comparisons in a  new process.  We don't do this now because
        # comparison function often use matplotlib, and matplotlib's Tkinter backend hangs when trying to create
        # a new figure in a new thread.
        # thread = Process(target = partial(func, records))
        # thread.start()
        # thread.join()

        func(records)
        _warn_with_prompt(use_prompt=False)

    def displayformat(self, new_format):
        assert new_format in ('nested', 'flat'), "Display format must be 'nested' or 'flat', not '{}'".format(new_format)
        self.display_format = new_format
        return ExperimentBrowser.REFRESH

    def errortrace(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint("Error Traces:", show_line=True, show_end=True):
            for record in records:
                with IndentPrint(record.get_id(), show_line=True):
                    print(record.get_error_trace())
        _warn_with_prompt(use_prompt=False)

    def delete(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print('{} out of {} Records will be deleted.'.format(len(records), sum(len(recs) for recs in self.exp_record_dict.values())))
        with IndentPrint():
            print(ExperimentRecordBrowser.get_record_table(records, ))
        response = input('Type "yes" to continue. >').strip().lower()
        if response == 'yes':
            clear_experiment_records([record.get_id() for record in records])
            print('Records deleted.')
            return ExperimentBrowser.REFRESH
        else:
            _warn_with_prompt('Records were not deleted.', use_prompt=False)

    def call(self, user_range):
        ids = select_experiments(user_range, self.exp_record_dict)
        for experiment_identifier in ids:
            load_experiment(experiment_identifier).call()

    def selectexp(self, user_range):
        exps_to_records = select_experiments(user_range, self.exp_record_dict, return_dict=True)
        with IndentPrint():
            print(self.get_experiment_list_str(exps_to_records))
        _warn_with_prompt('Experiment Selection "{}" includes {} out of {} experiments.'.format(user_range, len(exps_to_records), len(self.exp_record_dict)), use_prompt=not self.close_after)

    def selectrec(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        with IndentPrint():
            print(ExperimentRecordBrowser.get_record_table(records))
        _warn_with_prompt('Record Selection "{}" includes {} out of {} records.'.format(user_range, len(records), sum(len(recs) for recs in self.exp_record_dict.values())), use_prompt=not self.close_after)

    def side_by_side(self, user_range):
        records = select_experiment_records(user_range, self.exp_record_dict, flat=True)
        print(side_by_side([get_record_full_string(rec) for rec in records], max_linewidth=128))
        _warn_with_prompt(use_prompt=False)

    def argtable(self, *args):
        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', nargs = '?', default='all', help='A selection of experiment records to run.  Examples: "3" or "3-5", or "3,4,5"')
        args = parser.parse_args(args)
        records = select_experiment_records(args.user_range, self.exp_record_dict, flat=True)
        print_experiment_record_argtable(records)
        _warn_with_prompt(use_prompt=False)

    def records(self, ):
        browse_experiment_records(self.exp_record_dict.keys())

    def pull(self, *args):
        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', help='A selection of experiments whose records to pull.  Examples: "3" or "3-5", or "3,4,5"')
        parser.add_argument('machine_name', action='store', nargs = '?', default='all', help='Name of machine to pull from (must be listed in ~/.artemisrc)')
        # Following -p thing is temporary until we figure out how to deal with password only if needed
        parser.add_argument('-p', '--need_password', action='store_true', default=False, help='Put this flag if you need a password (leave it out if you have keys set up)')
        args = parser.parse_args(args)
        from artemis.remote.remote_machines import get_remote_machine_info
        info = get_remote_machine_info(args.machine_name)
        exp_names = select_experiments(args.user_range, self.exp_record_dict)
        output = pull_experiments(user=info['username'], ip=info['ip'], experiment_names=exp_names, include_variants=False, need_pass=args.need_password)
        print(output)
        return ExperimentBrowser.REFRESH

    def kill(self, *args):
        parser = argparse.ArgumentParser()
        parser.add_argument('user_range', action='store', help='A selection of experiments whose records to pull.  Examples: "3" or "3-5", or "3,4,5"')
        parser.add_argument('-s', '--skip', action='store_true', default=True, help='Skip the check that all selected records are currently running (just filter running ones)')
        args = parser.parse_args(args)

        records = select_experiment_records(args.user_range, self.exp_record_dict, flat=True)

        if not args.skip and (not all(record.get_status() is ExpStatusOptions.STARTED for record in records)):
            raise RecordSelectionError('Not all records you selected to kill were running: \n- {}'.format('\n- '.join('{}: {}'.format(rec.get_id(), rec.get_status()) for rec in records)))
        elif args.skip:
            records = [rec for rec in records if rec.get_status() is ExpStatusOptions.STARTED]

        if len(records)==0:
            raise RecordSelectionError('Selection "{}" selected no active processes to kill'.format(args.user_range))

        print('{} Running Experiments will be killed.'.format(len(records)))
        with IndentPrint():
            print(ExperimentRecordBrowser.get_record_table(records, ))
        response = input('Type "yes" to continue. >').strip().lower()
        if response == 'yes':
            for rec in records:
                rec.kill()
            print('Experiments killed.')
            return ExperimentBrowser.REFRESH
        else:
            _warn_with_prompt('Experiments were not killed.', use_prompt=False)

    def filter(self, user_range):
        self._filter = user_range if user_range not in ('-', '--clear') else None
        return ExperimentBrowser.REFRESH

    def filterrec(self, user_range):
        self._filterrec = user_range if user_range not in ('-', '--clear') else None
        return ExperimentBrowser.REFRESH

    def view(self, mode):
        self.view_mode = mode
        return ExperimentBrowser.REFRESH

    def explist(self, surround = ""):
        print("\n".join([surround+k+surround for k in self.exp_record_dict.keys()]))
        _warn_with_prompt(use_prompt=False)

    def quit(self):
        return ExperimentBrowser.QUIT


class ExpRecordDisplayFields(Enum):
    RUNS = 'Start Time'
    DURATION = 'Duration'
    STATUS = 'Status'
    ARGS_CHANGED = 'Args Changed?'
    RESULT_STR = 'Result'
    NOTES = 'Notes'


def _show_notes(rec):
    if rec.info.has_field(ExpInfoFields.NOTES):
        notes = rec.info.get_field(ExpInfoFields.NOTES)
        if notes is None:
            return ''
        else:
            return ';'.join(rec.info.get_field(ExpInfoFields.NOTES)).replace('\n', ';;')
    else:
        return ''


class _DisplaySettings(object):

    SETTINGS = {'ignore_valid_keys', ()}

    def __init__(self, settings_dict):
        self.settings_dict = settings_dict

    def __enter__(self):
        self.old_settings = _DisplaySettings.SETTINGS
        _DisplaySettings.SETTINGS = self.settings_dict

    def __exit__(self, exc_type, exc_val, exc_tb):
        _DisplaySettings.SETTINGS = self.old_settings

    @classmethod
    def get_setting(cls, name):
        return _DisplaySettings.SETTINGS[name]


_exp_record_field_getters = {
    ExpRecordDisplayFields.RUNS: lambda rec: format_time_stamp(rec.info.get_field(ExpInfoFields.TIMESTAMP)),
    ExpRecordDisplayFields.DURATION: lambda rec: format_duration(rec.info.get_field(ExpInfoFields.RUNTIME)) if rec.info.has_field(ExpInfoFields.RUNTIME) else '-',
    ExpRecordDisplayFields.ARGS_CHANGED: lambda rec: get_record_invalid_arg_string(rec, ignore_valid_keys=_DisplaySettings.get_setting('ignore_valid_keys')),
    ExpRecordDisplayFields.RESULT_STR: get_oneline_result_string,
    ExpRecordDisplayFields.STATUS: lambda rec: rec.info.get_field_text(ExpInfoFields.STATUS),
    ExpRecordDisplayFields.NOTES: _show_notes
}


def _get_record_rows(record_id, headers, raise_display_errors, truncate_to, ignore_valid_keys = ()):
    rec = load_experiment_record(record_id)

    with _DisplaySettings(dict(ignore_valid_keys=ignore_valid_keys)):
        if not raise_display_errors:
            values = []
            for h in headers:
                try:
                    values.append(_exp_record_field_getters[h](rec))
                except:
                    values.append('<Display Error>')
        else:
            values = [_exp_record_field_getters[h](rec) for h in headers]

    if truncate_to is not None:
        values = [truncate_string(val, truncation=truncate_to, message='...') for val in values]

    return values


def clear_ui_cache():
    shutil.rmtree(get_artemis_data_path('_ui_cache/'))


def _get_record_rows_cached(record_id, headers, raise_display_errors, truncate_to, ignore_valid_keys = ()):
    """
    We want to load the saved row only if:
    - The record is complete
    -
    :param record_id:
    :param headers:
    :return:
    """
    cache_key = compute_fixed_hash((record_id, [h.value for h in headers], truncate_to, ignore_valid_keys))
    path = get_artemis_data_path(os.path.join('_ui_cache', cache_key), make_local_dir=True)
    if os.path.exists(path):
        try:
            with open(path, 'rb') as f:
                record_rows = pickle.load(f)
            if len(record_rows)!=len(headers):
                os.remove(path)  # This should never happen.  But in case it somehow does, we just go ahead and compute again.
            else:
                return record_rows
        except:
            logging.warn('Failed to load cached record info: {}'.format(record_id))

    info_plus_status = _get_record_rows(record_id=record_id, headers=headers+[ExpRecordDisplayFields.STATUS],
                                        raise_display_errors=raise_display_errors, truncate_to=truncate_to, ignore_valid_keys=ignore_valid_keys)
    record_rows, status = info_plus_status[:-1], info_plus_status[-1]
    if status == ExpStatusOptions.STARTED:  # In this case it's still running (maybe) and we don't want to cache because it'll change
        return record_rows
    else:
        with open(path, 'wb') as f:
            pickle.dump(record_rows, f)
        return record_rows


def browse_experiment_records(*args, **kwargs):
    """
    Browse through experiment records.

    :param names: Filter by names of experiments
    :param filter_text: Filter by regular expression
    :return:
    """

    experiment_record_browser = ExperimentRecordBrowser(*args, **kwargs)
    experiment_record_browser.launch()


class ExperimentRecordBrowser(object):

    QUIT = 'Quit'
    HELP_TEXT = """
    q:                  Quit
    r:                  Refresh
    filter <text>       filter experiments
    viewfilters         View all filters on these results
    showall:            Show all experiments ever
    allnames:           Remove any name filters
    show <number>       Show experiment with number
    side_by_side 4,6,9       Compare experiments by their numbers.
    clearall            Delete all experements from your computer
"""

    def __init__(self, experiment_names = None, filter_text = None, raise_display_errors=False):
        """
        Browse through experiment records.

        :param names: Filter by names of experiments
        :param filter_text: Filter by regular expression
        :return:
        """
        self.experiment_names = experiment_names
        self.filters = [filter_text]
        self.record_ids = self.reload_ids()
        self.raise_display_errors = raise_display_errors

    def reload_ids(self):
        return get_all_record_ids(experiment_ids= self.experiment_names, filters=self.filters)

    @staticmethod
    def get_record_table(records = None, headers = ('#', 'Identifier', 'Start Time', 'Duration', 'Status', 'Valid', 'Notes', 'Result'), raise_display_errors = False, result_truncation=100):

        d = {
            '#': lambda: i,
            'Identifier': lambda: experiment_record.get_id(),
            'Start Time': lambda: experiment_record.info.get_field_text(ExpInfoFields.TIMESTAMP, replacement_if_none='?'),
            'Duration': lambda: experiment_record.info.get_field_text(ExpInfoFields.RUNTIME, replacement_if_none='?'),
            'Status': lambda: experiment_record.info.get_field_text(ExpInfoFields.STATUS, replacement_if_none='?'),
            'Args': lambda: experiment_record.info.get_field_text(ExpInfoFields.ARGS, replacement_if_none='?'),
            'Valid': lambda: get_record_invalid_arg_string(experiment_record, note_version='short'),
            'Notes': lambda: experiment_record.info.get_field_text(ExpInfoFields.NOTES, replacement_if_none='?'),
            'Result': lambda: get_oneline_result_string(experiment_record, truncate_to=128)
            # experiment_record.get_experiment().get_oneline_result_string(truncate_to=result_truncation) if is_experiment_loadable(experiment_record.get_experiment_id()) else '<Experiment not loaded>'
            }

        def get_col_info(headers):
            info = []
            for h in headers:
                try:
                    info.append(d[h]())
                except:
                    info.append('<Error displaying info>')
                    if raise_display_errors:
                        raise
            return info

        rows = []
        for i, experiment_record in enumerate(records):
            rows.append(get_col_info(headers))
        assert all_equal([len(headers)] + [len(row) for row in rows]), 'Header length: {}, Row Lengths: \n {}'.format(len(headers), [len(row) for row in rows])
        return tabulate(rows, headers=headers)

    def launch(self):

        func_lookup = {
            'q': self.quit,
            'h': self.help,
            'filter': self.filter,
            'showall': self.showall,
            'args': self.args,
            'rmfilters': self.rmfilters,
            'viewfilters': self.viewfilters,
            'side_by_side': self.compare,
            'show': self.show,
            'search': self.search,
            'delete': self.delete,
        }

        while True:

            print("=============== Experiment Records ==================")
            # print tabulate([[i]+e.get_row() for i, e in enumerate(record_infos)], headers=['#']+_ExperimentInfo.get_headers())
            print(self.get_record_table([load_experiment_record(rid) for rid in self.record_ids], raise_display_errors = self.raise_display_errors))
            print('-----------------------------------------------------')

            if self.experiment_names is not None or len(self.filters) != 0:
                print('Not showing all experiments.  Type "showall" to see all experiments, or "viewfilters" to view current filters.')
            user_input = input('Enter Command (show # to show and experiment, or h for help) >>').strip()
            parts = shlex.split(user_input)
            if len(parts)==0:
                print("You need to specify an command.  Press h for help.")
                continue
            cmd = parts[0]
            args = parts[1:]

            try:
                if cmd not in func_lookup:
                    raise _warn_with_prompt('Unknown Command: {}'.format(cmd))
                else:
                    return_val = func_lookup[cmd](*args)
                    if return_val==self.QUIT:
                        break
            except Exception as e:
                res = input('%s: %s\nEnter "e" to view the stacktrace, or anything else to continue.' % (e.__class__.__name__, e.message))
                if res.strip().lower() == 'e':
                    raise

    def _select_records(self, user_range):
        return select_experiment_records_from_list(user_range, self.record_ids)

    def quit(self):
        return self.QUIT

    def help(self):
        _warn_with_prompt(self.HELP_TEXT)

    def filter(self, filter_text):
        self.filters.append(filter_text)
        self.record_ids = self.reload_ids()

    def showall(self):
        self.filters = []
        self.experiment_names = None
        self.record_ids = self.reload_ids()

    def args(self, user_range):
        print(self.get_record_table(self._select_records(user_range), headers=['Identifier', 'Args']))

    def rmfilters(self):
        self.filters = []
        self.record_ids = self.reload_ids()

    def viewfilters(self):
        _warn_with_prompt('Filtering for: \n  Names in {}\n  Expressions: {}'.format(self.experiment_names, self.filters))

    def compare(self, user_range):
        identifiers = self._select_records(user_range)
        print_experiment_record_argtable(identifiers)
        _warn_with_prompt('')

    def show(self, user_range):
        record_ids = self._select_records(user_range)
        show_multiple_records([load_experiment_record(rid) for rid in record_ids])
        _warn_with_prompt('')

    def search(self, filter_text):
        print('Found the following Records: ')
        print(self.get_record_table([rid for rid in self.record_ids if filter_text in rid]))
        _warn_with_prompt()

    def delete(self, user_range):
        ids = self._select_records(user_range)
        print('We will delete the following experiments:')
        print(self.get_record_table(ids))
        conf = input("Going to clear {} of {} experiment records shown above.  Enter 'yes' to confirm: ".format(len(ids), len(self.record_ids)))
        if conf.strip().lower() == 'yes':
            clear_experiment_records(ids=ids)
        else:
            _warn_with_prompt("Did not delete experiments")
        self.record_ids = self.reload_ids()


if __name__ == '__main__':
    browse_experiment_records()




from artemis.experiments.experiment_record import *

import logging

logging.getLogger(__name__).warning('The module artemis.fileman.experiment_record is deprecated and will eventually be removed.  Import artemis.experiments.experiment_record instead.')



import sys
from artemis.general.local_capture import execute_and_capture_locals

__author__ = 'peter'


def test_execute_and_capture_locals():

    def func():
        a=3
        return a+4

    def outer_func():
        b=4
        def nested():
            return func()+b+5
        return nested

    out, local_vars = execute_and_capture_locals(func)
    assert out == 3+4
    assert local_vars == {'a': 3}
    assert sys.getprofile() is None

    out, local_vars = execute_and_capture_locals(outer_func())
    assert out == 7+4+5
    assert local_vars == {'b': 4, 'func': func}
    assert sys.getprofile() is None


if __name__ == '__main__':

    test_execute_and_capture_locals()


from artemis.fileman.file_getter import get_file
import re

__author__ = 'peter'


def read_book(code, max_characters = None):

    text = {
        'bible': read_the_bible,
        'fifty_shades_of_grey': read_fifty_shades_of_grey
        }[code](max_characters=max_characters)
    return text


def read_the_bible(max_characters = None):
    """
    Returns the King James Bible as a single string.
    Thanks to Janel (http://janelwashere.com/pages/bible_daily_reading.html) for compiling it.
    :param max_characters: You have the option to truncate it to a length of max_characters
        (If you're Jewish, for instance)
    :return: A string.
    """

    filename = get_file(
        relative_name = 'data/king_james_bible.txt',
        url = 'http://janelwashere.com/files/bible_daily.txt',
        )

    with open(filename) as f:
        text = f.read(-1 if max_characters is None else max_characters)

    return text


def read_fifty_shades_of_grey(max_characters = None):
    """
    Returns Fifty Shades of Gray, by EL James.
    :param max_characters:
    :return:
    """

    filename = get_file(
        relative_name = 'data/fifty_shades_of_grey.txt',
        url = None
        )

    with open(filename) as f:
        text = f.read(-1 if max_characters is None else max_characters)

    # Need to remove some weird non-ascii stuff.
    # http://stackoverflow.com/questions/20078816/replace-non-ascii-characters-with-a-single-space
    text = re.sub(r'[^\x00-\x7F]+',' ', text)

    return text


if __name__ == '__main__':

    book = 'fifty_shades_of_grey'
    n_characters = 10000

    print(read_book(book, n_characters))


import sys
from contextlib import contextmanager

__author__ = 'peter'

"""
A global shared "test_mode".  Allows all functions to modify their behaviour
in "test_mode".  Test mode should be set whenever the variable is set or when
things are beign run from pytest.

We do this to replace the previous solution of passing around a "test_mode" argumet
everywhere.

This is used in conjunction with conftest.py, which adds the _called_from_test flag.
"""

_TEST_MODE = False


def is_test_mode():
    return _TEST_MODE or (hasattr(sys, '_called_from_test') and sys._called_from_test)


def set_test_mode(state):
    global _TEST_MODE
    _TEST_MODE = state


@contextmanager
def hold_test_mode(test_mode = True):
    """
    Execute a block of code in test mode.  That is, any call to "is_test_mode" within that
    block will return True.  Usage:

    with hold_test_mode():
        if is_test_mode():
            dataset = get_my_big_dataset(n_samples = 10)  # Shorten the dataset just to run test
            n_epochs = 2
        else:
            dataset = get_my_big_dataset()
            n_epochs = 20
        ...
    """

    old_test_mode = _TEST_MODE
    set_test_mode(test_mode)
    yield
    set_test_mode(old_test_mode)


class UseTestContext(object):
    """
    DEPRECATED.... use hold_test_mode
    """

    def __init__(self, new_test_mode=True):
        self.new_test_mode = new_test_mode

    def __enter__(self):
        self._old_test_mode = _TEST_MODE
        set_test_mode(self.new_test_mode)

    def __exit__(self, exc_type, exc_val, exc_tb):
        set_test_mode(self._old_test_mode)



from numpy.random.mtrand import RandomState
import numpy as np
__author__ = 'peter'


def get_rng(initializer):
    """
    Creates a random number generator from either a seed, the time, or another random number generator.
    This is useful when you need to initialize random states on many levels, and you want the choice between
    specifying a seed directly or passing in an RNG from somewhere else.

    :param initializer: Can be:
        None, in which case the RNG is seeded on the time
        an int, in which case this is used as the seed
        a numpy RandomState object, in which case it's just passed through.
    :return: A numpy RandomState object.
    """
    if initializer is None or isinstance(initializer, int):
        return np.random.RandomState(initializer)
    elif isinstance(initializer, RandomState):
        return initializer


def array_info(arr):
    """
    :param arr: Return info about an array
    :return:
    """
    return '<{dtype} array of shape {shape}: mean: {mean:.3g}, std: {std:.3g}, min: {min:.3g}, max: {max:.3g} at {addr:s}>'.format(
        dtype = arr.dtype,
        shape = arr.shape,
        mean = arr.mean(),
        std = arr.std(),
        min = arr.min(),
        max = arr.max(),
        addr = hex(id(arr))
        )


def argtopk(x, k, axis=-1):
    return np.take(np.argpartition(-x, axis=axis, kth=k), np.arange(k), axis=axis)


def fast_array(objects):
    """
    np.ndarray does some checking on the list of objects first.  This just trusts that they're all similat
    :param objects:
    :return:
    """
    arr = np.empty((len(objects), )+objects[0].shape, dtype=objects[0].dtype)
    for i, ob in enumerate(objects):
        arr[i] = ob
    return arr


import numpy as np
from artemis.general.mymath import softmax, sigm

from artemis.ml.datasets.datasets import DataSet, DataCollection


__author__ = 'peter'


def get_synthethic_linear_dataset(noise_level = 0.1, n_input_dims = 20, n_output_dims = 4, n_training_samples = 1000,
        n_test_samples = 200, nonlinearity = None, offset_mag = 0, seed = 8158):
    """
    A Synthethic dataset that can be used for testing generalized linear models.

    :param noise_level:
    :param n_input_dims:
    :param n_output_dims:
    :param n_training_samples:
    :param n_test_samples:
    :param nonlinearity:
    :param seed:
    :return:
    """

    input_singleton = n_input_dims == 0
    if input_singleton:
        n_input_dims = 1

    output_singleton = n_output_dims == 0
    if output_singleton:  # Unfortunately we have to deal with the inconsistencies in numpy's handling of singleton dimensions.
        n_output_dims = 1

    rng = np.random.RandomState(seed)
    w = rng.randn(n_input_dims, n_output_dims) * 1/np.sqrt(n_input_dims)
    input_data = rng.randn(n_training_samples+n_test_samples, n_input_dims)
    target_data = np.dot(input_data, w) + offset_mag * rng.randn(n_output_dims) + noise_level*rng.randn(n_training_samples+n_test_samples, n_output_dims)
    if nonlinearity=='softmax':
        target_data = softmax(target_data, axis=1),
    elif nonlinearity=='sigmoid':
        target_data = sigm(target_data)
    elif nonlinearity=='argmax':
        target_data==np.argmax(target_data, axis=1)
    elif nonlinearity is None:
        target_data = target_data
    else:
        assert callable(nonlinearity), 'Unknown nonlinearity: {}'.format(nonlinearity)
        target_data = nonlinearity(target_data)

    if input_singleton:
        input_data = input_data[:, 0]

    if output_singleton:
        target_data = target_data[:, 0]

    return DataSet(
        training_set = DataCollection(input_data[:n_training_samples], target_data[:n_training_samples]),
        test_set = DataCollection(input_data[n_training_samples:], target_data[n_training_samples:]),
        )



from artemis.general.mymath import softmax
from artemis.ml.predictors.i_predictor import IPredictor
import numpy as np


class LogisticRegressor(IPredictor):

    def __init__(self, w, learning_rate=0.1):
        self.w = w
        self.learning_rate = learning_rate

    def predict(self, x):
        """
        :param x: An (n_samples, n_inputs) input
        :return: An (n_samples, n_classes) class probability
        """
        return softmax(x.dot(self.w), axis=1)

    def train(self, x, y):
        """
        :param x: An (n_samples, n_inputs) input
        :param y: An integer class label
        :return:
        """
        probs = self.predict(x)
        d_l_d_u = probs - onehot(y, n_classes=self.w.shape[1])
        self.w -= self.learning_rate * x.T.dot(d_l_d_u)

    @staticmethod
    def from_init(n_in, n_out, **kwargs):
        return LogisticRegressor(np.zeros((n_in, n_out)), **kwargs)


def onehot(labels, n_classes):
    """
    Turn a vector of labels into a onehot-encoding.
    :param labels:
    :param n_classes:
    :return:
    """
    values = np.zeros((len(labels), n_classes))
    values[np.arange(len(values)), labels] = 1
    return values